# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple, Union, Optional
from mrmustard.types import Matrix
from mrmustard.utils.parametrized import Parametrized
from mrmustard.lab.abstract import State, FockMeasurement
from mrmustard.lab.states import DisplacedSqueezed, Coherent
from mrmustard import settings
from mrmustard.math import Math

math = Math()

__all__ = ["PNRDetector", "ThresholdDetector", "Homodyne", "Heterodyne"]


class PNRDetector(Parametrized, FockMeasurement):
    r"""Photon Number Resolving detector.

    If ``len(modes) > 1`` the detector is applied in parallel to all of the modes provided.
    If a parameter is a single float, the parallel instances of the detector share that parameter.

    To apply mode-specific parmeters use a list of floats. The number of modes is determined (in order of priority)
    by the modes parameter, the cutoffs parameter, or the length of the efficiency and dark counts parameters.

    One can optionally set bounds for each parameter, which the optimizer will respect.

    It can be supplied the full stochastic channel, or it will compute it from
    the quantum efficiency (binomial) and the dark count probability (possonian).

    Args:
        efficiency (float or List[float]): list of quantum efficiencies for each detector
        efficiency_trainable (bool): whether the efficiency is trainable
        efficiency_bounds (Tuple[float, float]): bounds for the efficiency
        dark_counts (float or List[float]): list of expected dark counts
        dark_counts_trainable (bool): whether the dark counts are trainable
        dark_counts_bounds (Tuple[float, float]): bounds for the dark counts
        stochastic_channel (Optional 2d array): if supplied, this stochastic_channel will be used for belief propagation
        modes (Optional List[int]): list of modes to apply the detector to
        cutoffs (int or List[int]): largest phton number measurement cutoff for each mode
    """

    def __init__(
        self,
        efficiency: Union[float, List[float]] = 1.0,
        dark_counts: Union[float, List[float]] = 0.0,
        efficiency_trainable: bool = False,
        dark_counts_trainable: bool = False,
        efficiency_bounds: Tuple[Optional[float], Optional[float]] = (0.0, 1.0),
        dark_counts_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        stochastic_channel: Matrix = None,
        modes: List[int] = None,
        cutoffs: Union[int, List[int]] = None,
    ):
        if modes is not None:
            num_modes = len(modes)
        elif cutoffs is not None:
            num_modes = len(cutoffs)
        else:
            num_modes = max(len(math.atleast_1d(efficiency)), len(math.atleast_1d(dark_counts)))

        Parametrized.__init__(
            self,
            efficiency=math.atleast_1d(efficiency),
            dark_counts=math.atleast_1d(dark_counts),
            efficiency_trainable=efficiency_trainable,
            dark_counts_trainable=dark_counts_trainable,
            efficiency_bounds=efficiency_bounds,
            dark_counts_bounds=dark_counts_bounds,
            stochastic_channel=stochastic_channel,
            modes=modes if modes is not None else list(range(num_modes)),
            cutoffs=cutoffs if cutoffs is not None else [settings.PNR_INTERNAL_CUTOFF] * num_modes,
        )

        self.recompute_stochastic_channel()

    def should_recompute_stochastic_channel(self):
        return self._efficiency_trainable or self._dark_counts_trainable

    def recompute_stochastic_channel(self, cutoffs: List[int] = None):
        if cutoffs is None:
            cutoffs = [settings.PNR_INTERNAL_CUTOFF] * len(self._modes)
        self._internal_stochastic_channel = []
        if self._stochastic_channel is not None:
            self._internal_stochastic_channel = self._stochastic_channel
        else:
            efficiency = (
                math.tile(math.atleast_1d(self.efficiency), [len(cutoffs)])
                if len(math.atleast_1d(self.efficiency)) == 1
                else self.efficiency
            )
            dark_counts = (
                math.tile(math.atleast_1d(self.dark_counts), [len(cutoffs)])
                if len(math.atleast_1d(self.dark_counts)) == 1
                else self.dark_counts
            )
            for c, qe, dc in zip(cutoffs, efficiency, dark_counts):
                dark_prior = math.poisson(max_k=settings.PNR_INTERNAL_CUTOFF, rate=dc)
                condprob = math.binomial_conditional_prob(
                    success_prob=qe, dim_in=settings.PNR_INTERNAL_CUTOFF, dim_out=c
                )
                self._internal_stochastic_channel.append(
                    math.convolve_probs_1d(
                        condprob, [dark_prior, math.eye(settings.PNR_INTERNAL_CUTOFF)[0]]
                    )
                )


class ThresholdDetector(Parametrized, FockMeasurement):
    r"""Threshold detector: any Fock component other than vacuum counts toward a click in the detector.

    If ``len(modes) > 1`` the detector is applied in parallel to all of the modes provided.

    If a parameter is a single float, its value is applied to all of the parallel instances of the detector.

    To apply mode-specific values use a list of floats.

    It can be supplied the full conditional detection probabilities, or it will compute them from
    the quantum efficiency (binomial) and the dark count probability (bernoulli).

    Args:
        efficiency (float or List[float]): list of quantum efficiencies for each detector
        dark_count_prob (float or List[float]): list of dark count probabilities for each detector
        efficiency_trainable (bool): whether the efficiency is trainable
        dark_count_prob_trainable (bool): whether the dark count probabilities are trainable
        efficiency_bounds (Tuple[float, float]): bounds for the efficiency
        dark_count_prob_bounds (Tuple[float, float]): bounds for the dark count probabilities
        stochastic_channel (Optional 2d array): if supplied, this stochastic_channel will be used for belief propagation
        modes (Optional List[int]): list of modes to apply the detector to
    """

    def __init__(
        self,
        efficiency: Union[float, List[float]] = 1.0,
        dark_count_prob: Union[float, List[float]] = 0.0,
        efficiency_trainable: bool = False,
        dark_count_prob_trainable: bool = False,
        efficiency_bounds: Tuple[Optional[float], Optional[float]] = (0.0, 1.0),
        dark_count_prob_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        stochastic_channel=None,
        modes: List[int] = None,
    ):
        if modes is not None:
            num_modes = len(modes)
        else:
            num_modes = max(len(math.atleast_1d(efficiency)), len(math.atleast_1d(dark_count_prob)))
        if len(math.atleast_1d(efficiency)) == 1 and num_modes > 1:
            efficiency = math.tile(math.atleast_1d(efficiency), [num_modes])
        if len(math.atleast_1d(dark_count_prob)) == 1 and num_modes > 1:
            dark_count_prob = math.tile(math.atleast_1d(dark_count_prob), [num_modes])
        Parametrized.__init__(
            self,
            efficiency=efficiency,
            dark_count_prob=dark_count_prob,
            efficiency_trainable=efficiency_trainable,
            dark_count_prob_trainable=dark_count_prob_trainable,
            efficiency_bounds=efficiency_bounds,
            dark_count_prob_bounds=dark_count_prob_bounds,
            stochastic_channel=stochastic_channel,
            modes=modes or list(range(num_modes)),
            cutoffs=[2] * num_modes,
        )

        self.recompute_stochastic_channel()

    def should_recompute_stochastic_channel(self):
        return self._efficiency_trainable or self._dark_count_prob_trainable

    def recompute_stochastic_channel(self, cutoffs: List[int] = None):
        if cutoffs is None:
            cutoffs = [settings.PNR_INTERNAL_CUTOFF] * len(self._modes)
        self._internal_stochastic_channel = []
        if self._stochastic_channel is not None:
            self._internal_stochastic_channel = self._stochastic_channel
        else:
            for cut, qe, dc in zip(
                cutoffs,
                math.atleast_1d(self.efficiency)[:],
                math.atleast_1d(self.dark_count_prob)[:],
            ):
                row1 = math.pow(1.0 - qe, math.arange(cut)[None, :]) - math.cast(
                    dc, self.efficiency.dtype
                )
                row2 = 1.0 - row1
                # rest = math.zeros((cut - 2, cut), dtype=row1.dtype)
                condprob = math.concat([row1, row2], axis=0)
                self._internal_stochastic_channel.append(condprob)


class Heterodyne(Parametrized, State):
    r"""Heterodyne measurement on given modes."""

    def __new__(
        cls,
        x: Union[float, List[float]] = 0.0,
        y: Union[float, List[float]] = 0.0,
        modes: List[int] = None,
    ):
        instance = Coherent(x=x, y=y, modes=modes)
        instance.__class__ = cls  # NOTE: naughty?
        return instance

    def __init__(self, *args, **kwargs):
        pass


class Homodyne(Parametrized, State):
    r"""Heterodyne measurement on given modes."""

    def __new__(
        cls,
        quadrature_angle: Union[float, List[float]],
        result: Union[float, List[float]] = 0.0,
        modes: List[int] = None,
        r: Union[float, List[float]] = None,
    ):
        quadrature_angle = math.astensor(quadrature_angle, dtype="float64")
        result = math.astensor(result, dtype="float64")
        x = result * math.cos(quadrature_angle)
        y = result * math.sin(quadrature_angle)
        instance = DisplacedSqueezed(
            r=settings.HOMODYNE_SQUEEZING if r is None else math.astensor(r, dtype="float64"),
            phi=2 * quadrature_angle,
            x=x,
            y=y,
            modes=modes,
        )
        instance.__class__ = cls
        return instance

    def __init__(self, *args, **kwargs):
        pass



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch
from thewalrus import hermite_multidimensional, grad_hermite_multidimensional

from .math_interface import MathInterface
from mrmustard.math.autocast import Autocast
from mrmustard.types import *


class TorchMath(MathInterface):

    float64 = torch.float64
    float32 = torch.float32
    complex64 = torch.complex64
    complex128 = torch.complex128

    def __getattr__(self, name):
        return getattr(torch, name)

    # ~~~~~~~~~
    # Basic ops
    # ~~~~~~~~~

    def atleast_1d(self, array: torch.Tensor, dtype=None) -> torch.Tensor:
        return self.cast(torch.reshape(self.astensor(array), [-1]), dtype)

    def astensor(self, array: Union[np.ndarray, torch.Tensor], dtype=None) -> torch.Tensor:
        return self.cast(torch.tensor(array), dtype)

    def conj(self, array: torch.Tensor) -> torch.Tensor:
        return torch.conj(array)

    def real(self, array: torch.Tensor) -> torch.Tensor:
        return torch.real(array)

    def imag(self, array: torch.Tensor) -> torch.Tensor:
        return torch.imag(array)

    def cos(self, array: torch.Tensor) -> torch.Tensor:
        return torch.cos(array)

    def cosh(self, array: torch.Tensor) -> torch.Tensor:
        return torch.cosh(array)

    def sinh(self, array: torch.Tensor) -> torch.Tensor:
        return torch.sinh(array)

    def sin(self, array: torch.Tensor) -> torch.Tensor:
        return torch.sin(array)

    def exp(self, array: torch.Tensor) -> torch.Tensor:
        return torch.exp(array)

    def sqrt(self, x: torch.Tensor, dtype=None) -> torch.Tensor:
        return self.cast(torch.sqrt(x), dtype)

    def lgamma(self, x: torch.Tensor) -> torch.Tensor:
        return torch.lgamma(x)

    def log(self, x: torch.Tensor) -> torch.Tensor:
        return torch.log(x)

    def cast(self, x: torch.Tensor, dtype=None) -> torch.Tensor:
        if dtype is None:
            return x
        return x.to(dtype)

    @Autocast()
    def maximum(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.maximum(a, b)

    @Autocast()
    def minimum(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.minimum(a, b)

    def abs(self, array: torch.Tensor) -> torch.Tensor:
        return torch.abs(array)

    def expm(self, matrix: torch.Tensor) -> torch.Tensor:
        return torch.matrix_exp(matrix)

    def norm(self, array: torch.Tensor) -> torch.Tensor:
        """Note that the norm preserves the type of array."""
        return torch.norm(array)

    @Autocast()
    def matmul(
        self,
        a: torch.Tensor,
        b: torch.Tensor,
        transpose_a=False,
        transpose_b=False,
        adjoint_a=False,
        adjoint_b=False,
    ) -> torch.Tensor:
        return torch.matmul(a, b)

    @Autocast()
    def matvec(
        self, a: torch.Tensor, b: torch.Tensor, transpose_a=False, adjoint_a=False
    ) -> torch.Tensor:
        return torch.mv(a, b)

    @Autocast()
    def tensordot(self, a: torch.Tensor, b: torch.Tensor, axes: List[int]) -> torch.Tensor:
        return torch.tensordot(a, b, axes)

    def einsum(self, string: str, *tensors) -> torch.Tensor:
        return torch.einsum(string, *tensors)

    def inv(self, a: torch.Tensor) -> torch.Tensor:
        return torch.inverse(a)

    def pinv(self, array: torch.Tensor) -> torch.Tensor:
        return torch.pinverse(array)

    def det(self, a: torch.Tensor) -> torch.Tensor:
        return torch.det(a)

    def tile(self, array: torch.Tensor, repeats: Sequence[int]) -> torch.Tensor:
        return torch.tile(array, repeats)

    def diag(self, array: torch.Tensor, k: int = 0) -> torch.Tensor:
        return torch.diag(array, k=k)

    def diag_part(self, array: torch.Tensor) -> torch.Tensor:
        return torch.diag_embed(array)

    def pad(
        self,
        array: torch.Tensor,
        paddings: Sequence[Tuple[int, int]],
        mode="constant",
        constant_values=0,
    ) -> torch.Tensor:
        return torch.nn.functional.pad(array, paddings, mode=mode, value=constant_values)

    @Autocast()
    def convolution(
        self,
        array: torch.Tensor,
        filters: torch.Tensor,
        strides: Optional[List[int]] = None,
        padding="VALID",
        data_format="NWC",
        dilations: Optional[List[int]] = None,
    ) -> torch.Tensor:
        r"""Wrapper for ``torch.nn.Conv1d`` and ``torch.nn.Conv2d``.

        Args:
            1D convolution: Tensor of shape [batch_size, input_channels, signal_length].
            2D convolution: [batch_size, input_channels, input_height, input_width]

        Returns:
        """

        batch_size = array.shape[0]
        input_channels = array.shape[1]
        output_channels = ...  # TODO: unsure of how to get output channels

        if array.dim() == 3:  # 1D case
            signal_length = array.shape[2]

            m = torch.nn.Conv1d(
                input_channels,
                output_channels,
                filters,
                stride=strides,
                padding=padding,
                dtype=data_format,
                dilation=dilations,
            )
            return m(array)
        elif array.dim() == 4:  # 2D case
            input_height = array.shape[2]
            input_width = array.shape[3]

            m = torch.nn.Conv2d(
                input_channels,
                output_channels,
                filters,
                stride=strides,
                padding=padding,
                dtype=data_format,
                dilation=dilations,
            )
            return m(array)
        else:
            raise NotImplementedError

    def transpose(self, a: torch.Tensor, perm: List[int] = None) -> torch.Tensor:
        if a is None:
            return None  # TODO: remove and address None inputs where transpose is used
        return torch.transpose(a, perm[0], perm[1])

    def reshape(self, array: torch.Tensor, shape: Sequence[int]) -> torch.Tensor:
        return torch.reshape(array, shape)

    def sum(self, array: torch.Tensor, axes: Sequence[int] = None):
        return torch.sum(array, axes)

    def arange(
        self, start: int, limit: int = None, delta: int = 1, dtype=torch.float64
    ) -> torch.Tensor:
        return torch.arange(start, limit, delta, dtype=dtype)

    @Autocast()
    def outer(self, array1: torch.Tensor, array2: torch.Tensor) -> torch.Tensor:
        return torch.tensordot(array1, array2, [[], []])

    def eye(self, size: int, dtype=torch.float64) -> torch.Tensor:
        return torch.eye(size, dtype=dtype)

    def zeros(self, shape: Sequence[int], dtype=torch.float64) -> torch.Tensor:
        return torch.zeros(shape, dtype=dtype)

    def zeros_like(self, array: torch.Tensor) -> torch.Tensor:
        return torch.zeros_like(array)

    def ones(self, shape: Sequence[int], dtype=torch.float64) -> torch.Tensor:
        return torch.ones(shape, dtype=dtype)

    def ones_like(self, array: torch.Tensor) -> torch.Tensor:
        return torch.ones_like(array)

    def gather(self, array: torch.Tensor, indices: torch.Tensor, axis: int = None) -> torch.Tensor:
        # TODO: gather works differently in Pytorch vs Tensorflow.

        return torch.gather(array, axis, indices)

    def trace(self, array: torch.Tensor, dtype=None) -> torch.Tensor:
        return self.cast(torch.trace(array), dtype)

    def concat(self, values: Sequence[torch.Tensor], axis: int) -> torch.Tensor:
        return torch.cat(values, axis)

    def update_tensor(
        self, tensor: torch.Tensor, indices: torch.Tensor, values: torch.Tensor, dims: int = 0
    ):
        # TODO: dims need to be an argument, or should be interpreted from the other data

        return tensor.scatter_(dims, indices, values)

    def update_add_tensor(
        self, tensor: torch.Tensor, indices: torch.Tensor, values: torch.Tensor, dims: int = 0
    ):
        # TODO: dims need to be an argument, or should be interpreted from the other data

        return tensor.scatter_add_(dims, indices, values)

    def constraint_func(
        self, bounds: Tuple[Optional[float], Optional[float]]
    ) -> Optional[Callable]:
        bounds = (
            -np.inf if bounds[0] is None else bounds[0],
            np.inf if bounds[1] is None else bounds[1],
        )
        if not bounds == (-np.inf, np.inf):
            constraint: Optional[Callable] = lambda x: torch.clamp(x, min=bounds[0], max=bounds[1])
        else:
            constraint = None
        return constraint

    def new_variable(
        self, value, bounds: Tuple[Optional[float], Optional[float]], name: str, dtype=torch.float64
    ):
        return torch.tensor(value, requires_grad=True)

    def new_constant(self, value, name: str, dtype=torch.float64):
        return torch.tensor(value)

    def asnumpy(self, tensor: torch.Tensor) -> Tensor:
        return tensor.numpy()

    def hash_tensor(self, tensor: torch.Tensor) -> str:
        return hash(tensor)

    def hermite_renormalized(
        self, A: torch.Tensor, B: torch.Tensor, C: torch.Tensor, shape: Tuple[int]
    ) -> torch.Tensor:  # TODO this is not ready
        r"""Renormalized multidimensional Hermite polynomial.

        This is given by the "exponential" Taylor series of :math:`exp(Ax^2 + Bx + C)` at zero,
        where the series has :math:`sqrt(n!)` at the denominator rather than `n!`.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            shape: The shape of the final tensor.

        Returns:
            The renormalized Hermite polynomial of given shape.
        """
        raise NotImplementedError

    def DefaultEuclideanOptimizer(self, params) -> torch.optim.Optimizer:
        r"""Default optimizer for the Euclidean parameters."""
        self.optimizer = torch.optim.Adam(params, lr=0.001)
        return self.optimizer

    def value_and_gradients(
        self, cost_fn: Callable, parameters: Dict[str, List[Trainable]]
    ) -> Tuple[torch.Tensor, Dict[str, List[torch.Tensor]]]:
        r"""Computes the loss and gradients of the given cost function.

        Args:
            cost_fn (Callable): The cost function. Takes in two arguments:
                - Output: The output tensor of the model.
            parameters (Dict): The parameters to optimize in three kinds:
                symplectic, orthogonal and euclidean.
            optimizer: The optimizer to be used by the math backend.

        Returns:
            The loss and the gradients.
        """
        self.optimizer.zero_grad()
        loss = (
            cost_fn()
        )  # TODO: I think this should be cost_fn(params), but if it works I think it is fine.
        loss.backward()
        self.optimizer.step()

        grads = [p.grad for p in parameters]

        return loss, grads

    def eigvals(self, tensor: torch.Tensor) -> Tensor:
        """Returns the eigenvalues of a matrix."""
        return torch.linalg.eigvals(tensor)

    def eigvalsh(self, tensor: torch.Tensor) -> Tensor:
        """Returns the eigenvalues of a Real Symmetric or Hermitian matrix."""
        return torch.linalg.eigvalsh(tensor)

    def svd(self, tensor: torch.Tensor) -> Tensor:
        """Returns the Singular Value Decomposition of a matrix."""
        return torch.linalg.svd(tensor)

    def xlogy(self, x: torch.Tensor, y: torch.Tensor) -> Tensor:
        """Returns 0 if ``x == 0``, and ``x * log(y)`` otherwise, elementwise."""
        return torch.xlogy(x, y)

    def sqrtm(self, tensor: torch.Tensor) -> Tensor:
        raise NotImplementedError

    def boolean_mask(self, tensor: torch.Tensor, mask: torch.Tensor) -> Tensor:
        return torch.masked_select(tensor, mask)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
from hypothesis import settings, Verbosity

print("pytest.conf -----------------------")

settings.register_profile("ci", max_examples=1000, deadline=None)
settings.register_profile("dev", max_examples=10, deadline=None)
settings.register_profile("debug", max_examples=10, verbosity=Verbosity.verbose, deadline=None)

settings.load_profile(os.getenv(u"HYPOTHESIS_PROFILE", "dev"))



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The physics module contains high-level functions for performing calculations on quantum objects.
It allows for objects with different representations to be used seamlessly in the same calculation, e.g., ``fidelity(A, B)`` works whether A and B are in the Fock or Gaussian representation or a mix of both.

All the functions are automatically differentiated and can be used in conjunction with an
optimization routine.
"""

from mrmustard.physics import fock, gaussian
from mrmustard import settings


def fidelity(A, B) -> float:
    r"""Calculates the fidelity between two quantum states.

    Args:
        A (State) The first quantum state.
        B (State) The second quantum state.

    Returns:
        float: The fidelity between the two states.
    """
    if A.is_gaussian and B.is_gaussian:
        return gaussian.fidelity(A.means, A.cov, B.means, B.cov, settings.HBAR)
    return fock.fidelity(A.fock, B.fock, a_ket=A._ket is not None, b_ket=B._ket is not None)


def normalize(A):
    r"""Returns the normalized state.

    Args:
        A (State): the quantum state

    Returns:
        State: the normalized state
    """
    if A.is_gaussian:
        A._norm = 1.0
        return A
    if A.is_mixed:
        return A.__class__(dm=fock.normalize(A.dm(), is_dm=True))
    else:
        return A.__class__(ket=fock.normalize(A.ket(), is_dm=False))


def norm(A) -> float:
    r"""Calculates the norm of a quantum state.

    The norm is equal to the trace of the density matrix if the state
    is mixed and to the norm of the state vector if the state is pure.

    Args:
        A (State): the quantum state

    Returns:
        float: the norm of the state
    """
    if A.is_gaussian:
        return A._norm
    return fock.norm(A.fock, is_dm=A.is_mixed)


def overlap(A, B) -> float:
    r"""Calculates the overlap between two quantum states.

    If the states are both pure it returns :math:`|<A|B>|^2`, if one is mixed it returns :math:`<A|B|A>`
    and if both are mixed it returns :math:`Tr[AB]`.

    Args:
        A (State) the first quantum state
        B (State) the second quantum state

    Returns:
        float: the overlap between the two states
    """
    raise NotImplementedError
    if A.is_gaussian and B.is_gaussian:
        return gaussian.overlap(A.means, A.cov, B.means, B.cov, settings.HBAR)
    return fock.overlap(A.fock, B.fock, a_dm=A.is_mixed, b_dm=B.is_mixed)


def von_neumann_entropy(A) -> float:
    r"""Calculates the Von Neumann entropy of a quantum state.

    Args:
        A (State) the quantum state

    Returns:
        float: the Von Neumann entropy of the state
    """
    if A.is_gaussian:
        return gaussian.von_neumann_entropy(A.cov, settings.HBAR)
    return fock.von_neumann_entropy(A.fock, a_dm=A.is_mixed)


def relative_entropy(A, B) -> float:
    r"""Calculates the relative entropy between two quantum states.

    Args:
        A (State) the first quantum state
        B (State) the second quantum state

    Returns:
        float: the relative entropy between the two states
    """
    raise NotImplementedError
    if A.is_gaussian and B.is_gaussian:
        return gaussian.relative_entropy(A.means, A.cov, B.means, B.cov, settings.HBAR)
    return fock.relative_entropy(A.fock, B.fock, a_dm=A.is_mixed, b_dm=B.is_mixed)


def trace_distance(A, B) -> float:
    r"""Calculates the trace distance between two quantum states.

    Args:
        A (State) the first quantum state
        B (State) the second quantum state

    Returns:
        float: the trace distance between the two states
    """
    raise NotImplementedError
    if A.is_gaussian and B.is_gaussian:
        return gaussian.trace_distance(A.means, A.cov, B.means, B.cov, settings.HBAR)
    return fock.trace_distance(A.fock, B.fock, a_dm=A.is_mixed, b_dm=B.is_mixed)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod, abstractproperty
from mrmustard.types import *
import numpy as np
from functools import lru_cache
from scipy.special import binom
from scipy.stats import unitary_group
from itertools import product


class MathInterface(ABC):
    r"""The interface that all backends must implement."""
    _euclidean_opt: type = None  # NOTE this is an object that

    # backend is a singleton
    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    @abstractmethod
    def __getattr__(self, name):
        ...  # pass the call to the actual backend

    # ~~~~~~~~~
    # Basic ops
    # ~~~~~~~~~

    @abstractmethod
    def abs(self, array: Tensor) -> Tensor:
        r"""Returns the absolute value of array.

        Args:
            array (array): array to take the absolute value of

        Returns:
            array: absolute value of array
        """
        ...

    @abstractmethod
    def any(self, array: Tensor) -> bool:
        r"""Returns ``True`` if any element of array is ``True``.

        Args:
            array (array): array to check

        Returns:
            bool: True if any element of array is True
        """
        ...

    @abstractmethod
    def arange(self, start: int, limit: int = None, delta: int = 1) -> Tensor:
        r"""Returns an array of evenly spaced values within a given interval.

        Args:
            start (int): start of the interval
            limit (int): end of the interval
            delta (int): step size

        Returns:
            array: array of evenly spaced values
        """
        ...  # NOTE: is float64 by default

    @abstractmethod
    def asnumpy(self, tensor: Tensor) -> Tensor:
        r"""Converts a tensor to a NumPy array.

        Args:
            tensor (array): tensor to convert

        Returns:
            array: NumPy array
        """
        ...

    @abstractmethod
    def assign(self, tensor: Tensor, value: Tensor) -> Tensor:
        r"""Assigns value to tensor.

        Args:
            tensor (array): tensor to assign to
            value (array): value to assign

        Returns:
            array: tensor with value assigned
        """
        ...

    @abstractmethod
    def astensor(self, array: Tensor, dtype: str) -> Tensor:
        r"""Converts a numpy array to a tensor.

        Args:
            array (array): numpy array to convert
            dtype (str): dtype of the tensor

        Returns:
            array: tensor with dtype
        """
        ...

    @abstractmethod
    def atleast_1d(self, array: Tensor, dtype=None) -> Tensor:
        r"""Returns an array with at least one dimension.

        Args:
            array (array): array to convert
            dtype (dtype): data type of the array

        Returns:
            array: array with at least one dimension
        """
        ...

    @abstractmethod
    def cast(self, array: Tensor, dtype) -> Tensor:
        r"""Casts ``array`` to ``dtype``.

        Args:
            array (array): array to cast
            dtype (dtype): data type to cast to

        Returns:
            array: array cast to dtype
        """
        ...

    @abstractmethod
    def clip(self, array: Tensor, a_min: float, a_max: float) -> Tensor:
        r"""Clips array to the interval ``[a_min, a_max]``.

        Args:
            array (array): array to clip
            a_min (float): minimum value
            a_max (float): maximum value

        Returns:
            array: clipped array
        """
        ...

    @abstractmethod
    def concat(self, values: Sequence[Tensor], axis: int) -> Tensor:
        r"""Concatenates values along the given axis.

        Args:
            values (array): values to concatenate
            axis (int): axis along which to concatenate

        Returns:
            array: concatenated values
        """
        ...

    @abstractmethod
    def conj(self, array: Tensor) -> Tensor:
        r"""Returns the complex conjugate of array.

        Args:
            array (array): array to take the complex conjugate of

        Returns:
            array: complex conjugate of array
        """
        ...

    @abstractmethod
    def constraint_func(
        self, bounds: Tuple[Optional[float], Optional[float]]
    ) -> Optional[Callable]:
        r"""Returns a constraint function for the given bounds.

        A constraint function will clip the value to the interval given by the bounds.

        .. note::

            The upper and/or lower bounds can be ``None``, in which case the constraint
            function will not clip the value.

        Args:
            bounds (tuple): bounds of the constraint

        Returns:
            function: constraint function
        """
        ...

    @abstractmethod
    def convolution(
        self,
        array: Tensor,
        filters: Tensor,
        padding="VALID",
        data_format="NWC",
    ) -> Tensor:  # TODO: remove strides and data_format?
        r"""Performs a convolution on array with filters.

        Args:
            array (array): array to convolve
            filters (array): filters to convolve with
            padding (str): padding mode
            data_format (str): data format of the array

        Returns:
            array: convolved array
        """
        ...

    @abstractmethod
    def cos(self, array: Tensor) -> Tensor:
        r"""Returns the cosine of array.

        Args:
            array (array): array to take the cosine of

        Returns:
            array: cosine of array
        """
        ...

    @abstractmethod
    def cosh(self, array: Tensor) -> Tensor:
        r"""Returns the hyperbolic cosine of array.

        Args:
            array (array): array to take the hyperbolic cosine of

        Returns:
            array: hyperbolic cosine of array
        """
        ...

    @abstractmethod
    def det(self, matrix: Tensor) -> Tensor:
        r"""Returns the determinant of matrix.

        Args:
            matrix (matrix): matrix to take the determinant of

        Returns:
            determinant of matrix
        """
        ...

    @abstractmethod
    def diag(self, array: Tensor, k: int) -> Tensor:
        r"""Returns the array made by inserting the given array along the :math:`k`-th diagonal.

        Args:
            array (array): array to insert
            k (int): kth diagonal to insert array into

        Returns:
            array: array with array inserted into the kth diagonal
        """
        ...

    @abstractmethod
    def diag_part(self, array: Tensor) -> Tensor:
        r"""Returns the array of the main diagonal of array.

        Args:
            array (array): array to extract the main diagonal of

        Returns:
            array: array of the main diagonal of array
        """
        ...

    @abstractmethod
    def einsum(self, string: str, *tensors) -> Tensor:
        r"""Returns the result of the Einstein summation convention on the tensors.

        Args:
            string (str): string of the Einstein summation convention
            tensors (array): tensors to perform the Einstein summation on

        Returns:
            array: result of the Einstein summation convention
        """
        ...

    @abstractmethod
    def exp(self, array: Tensor) -> Tensor:
        r"""Returns the exponential of array element-wise.

        Args:
            array (array): array to take the exponential of

        Returns:
            array: exponential of array
        """
        ...

    @abstractmethod
    def expand_dims(self, array: Tensor, axis: int) -> Tensor:
        r"""Returns the array with an additional dimension inserted at the given axis.

        Args:
            array (array): array to expand
            axis (int): axis to insert the new dimension

        Returns:
            array: array with an additional dimension inserted at the given axis
        """
        ...

    @abstractmethod
    def expm(self, matrix: Tensor) -> Tensor:
        r"""Returns the matrix exponential of matrix.

        Args:
            matrix (matrix): matrix to take the exponential of

        Returns:
            matrix: exponential of matrix
        """
        ...

    @abstractmethod
    def eye(self, size: int, dtype) -> Tensor:
        r"""Returns the identity matrix of size.

        Args:
            size (int): size of the identity matrix
            dtype (dtype): data type of the identity matrix

        Returns:
            matrix: identity matrix
        """
        ...

    @abstractmethod
    def from_backend(self, value: Any) -> bool:
        r"""Returns whether the given tensor is a tensor of the concrete backend."""
        ...

    @abstractmethod
    def gather(self, array: Tensor, indices: Tensor, axis: int) -> Tensor:
        r"""Returns the values of the array at the given indices.

        Args:
            array (array): array to gather values from
            indices (array): indices to gather values from
            axis (int): axis to gather values from

        Returns:
            array: values of the array at the given indices
        """
        ...

    @abstractmethod
    def hash_tensor(self, tensor: Tensor) -> int:
        r"""Returns the hash of the given tensor.

        Args:
            tensor (array): tensor to hash

        Returns:
            int: hash of the given tensor
        """
        ...

    @abstractmethod
    def hermite_renormalized(self, A: Matrix, B: Vector, C: Scalar, shape: Sequence[int]) -> Tensor:
        r"""Returns the array of hermite renormalized polynomials of the given coefficients.

        Args:
            A (array): Matrix coefficient of the hermite polynomial
            B (array): Vector coefficient of the hermite polynomial
            C (array): Scalar coefficient of the hermite polynomial
            shape (tuple): shape of the hermite polynomial

        Returns:
            array: renormalized hermite polynomials
        """
        ...

    @abstractmethod
    def imag(self, array: Tensor) -> Tensor:
        r"""Returns the imaginary part of array.

        Args:
            array (array): array to take the imaginary part of

        Returns:
            array: imaginary part of array
        """
        ...

    @abstractmethod
    def inv(self, tensor: Tensor) -> Tensor:
        r"""Returns the inverse of tensor.

        Args:
            tensor (array): tensor to take the inverse of

        Returns:
            array: inverse of tensor
        """
        ...

    @abstractmethod
    def is_trainable(self, tensor: Tensor) -> bool:
        r"""Returns whether the given tensor is trainable."""
        ...

    @abstractmethod
    def lgamma(self, x: Tensor) -> Tensor:
        r"""Returns the natural logarithm of the gamma function of ``x``.

        Args:
            x (array): array to take the natural logarithm of the gamma function of

        Returns:
            array: natural logarithm of the gamma function of ``x``
        """
        ...

    @abstractmethod
    def log(self, x: Tensor) -> Tensor:
        r"""Returns the natural logarithm of ``x``.

        Args:
            x (array): array to take the natural logarithm of

        Returns:
            array: natural logarithm of ``x``
        """
        ...

    @abstractmethod
    def matmul(
        self,
        a: Tensor,
        b: Tensor,
        transpose_a=False,
        transpose_b=False,
        adjoint_a=False,
        adjoint_b=False,
    ) -> Tensor:
        r"""Returns the matrix product of ``a`` and ``b``.

        Args:
            a (array): first matrix to multiply
            b (array): second matrix to multiply
            transpose_a (bool): whether to transpose ``a``
            transpose_b (bool): whether to transpose ``b``
            adjoint_a (bool): whether to adjoint ``a``
            adjoint_b (bool): whether to adjoint ``b``

        Returns:
            array: matrix product of ``a`` and ``b``
        """
        ...

    @abstractmethod
    def matvec(self, a: Matrix, b: Vector, transpose_a=False, adjoint_a=False) -> Tensor:
        r"""Returns the matrix vector product of ``a`` (matrix) and ``b`` (vector).

        Args:
            a (array): matrix to multiply
            b (array): vector to multiply
            transpose_a (bool): whether to transpose ``a``
            adjoint_a (bool): whether to adjoint ``a``

        Returns:
            array: matrix vector product of ``a`` and ``b``
        """
        ...

    @abstractmethod
    def maximum(self, a: Tensor, b: Tensor) -> Tensor:
        r"""Returns the element-wise maximum of ``a`` and ``b``.

        Args:
            a (array): first array to take the maximum of
            b (array): second array to take the maximum of

        Returns:
            array: element-wise maximum of ``a`` and ``b``
        """
        ...

    @abstractmethod
    def minimum(self, a: Tensor, b: Tensor) -> Tensor:
        r"""Returns the element-wise minimum of ``a`` and ``b``.

        Args:
            a (array): first array to take the minimum of
            b (array): second array to take the minimum of

        Returns:
            array: element-wise minimum of ``a`` and ``b``
        """
        ...

    @abstractmethod
    def new_variable(
        self, value: Tensor, bounds: Tuple[Optional[float], Optional[float]], name: str
    ) -> Tensor:
        r"""Returns a new variable with the given value and bounds.

        Args:
            value (array): value of the new variable
            bounds (tuple): bounds of the new variable
            name (str): name of the new variable

        Returns:
            array: new variable
        """
        ...

    @abstractmethod
    def new_constant(self, value: Tensor, name: str) -> Tensor:
        r"""Returns a new constant with the given value.

        Args:
            value (array): value of the new constant
            name (str): name of the new constant

        Returns:
            array: new constant
        """
        ...

    @abstractmethod
    def norm(self, array: Tensor) -> Tensor:
        r"""Returns the norm of array.

        Args:
            array (array): array to take the norm of

        Returns:
            array: norm of array
        """
        ...

    @abstractmethod
    def ones(self, shape: Sequence[int], dtype) -> Tensor:
        r"""Returns an array of ones with the given ``shape`` and ``dtype``.

        Args:
            shape (tuple): shape of the array
            dtype (type): dtype of the array

        Returns:
            array: array of ones
        """
        ...  # NOTE : should be float64 by default

    @abstractmethod
    def ones_like(self, array: Tensor) -> Tensor:
        r"""Returns an array of ones with the same shape and ``dtype`` as ``array``.

        Args:
            array (array): array to take the shape and dtype of

        Returns:
            array: array of ones
        """
        ...

    @abstractmethod
    def outer(self, array1: Tensor, array2: Tensor) -> Tensor:
        r"""Returns the outer product of ``array1`` and ``array2``.

        Args:
            array1 (array): first array to take the outer product of
            array2 (array): second array to take the outer product of

        Returns:
            array: outer product of array1 and array2
        """
        ...

    @abstractmethod
    def pad(
        self, array: Tensor, paddings: Sequence[Tuple[int, int]], mode="CONSTANT", constant_values=0
    ) -> Tensor:
        r"""Returns the padded array.

        Args:
            array (array): array to pad
            paddings (tuple): paddings to apply
            mode (str): mode to apply the padding
            constant_values (int): constant values to use for padding

        Returns:
            array: padded array
        """
        ...

    @abstractmethod
    def pinv(self, matrix: Tensor) -> Tensor:
        r"""Returns the pseudo-inverse of matrix.

        Args:
            matrix (array): matrix to take the pseudo-inverse of

        Returns:
            array: pseudo-inverse of matrix
        """
        ...

    @abstractmethod
    def pow(self, x: Tensor, y: Tensor) -> Tensor:
        r"""Returns :math:`x^y`. Broadcasts ``x`` and ``y`` if necessary.
        Args:
            x (array): base
            y (array): exponent

        Returns:
            array: :math:`x^y`
        """
        ...

    @abstractmethod
    def real(self, array: Tensor) -> Tensor:
        r"""Returns the real part of ``array``.

        Args:
            array (array): array to take the real part of

        Returns:
            array: real part of ``array``
        """
        ...

    @abstractmethod
    def reshape(self, array: Tensor, shape: Sequence[int]) -> Tensor:
        r"""Returns the reshaped array.

        Args:
            array (array): array to reshape
            shape (tuple): shape to reshape the array to

        Returns:
            array: reshaped array
        """
        ...

    @abstractmethod
    def sin(self, array: Tensor) -> Tensor:
        r"""Returns the sine of ``array``.

        Args:
            array (array): array to take the sine of

        Returns:
            array: sine of ``array``
        """
        ...

    @abstractmethod
    def sinh(self, array: Tensor) -> Tensor:
        r"""Returns the hyperbolic sine of ``array``.

        Args:
            array (array): array to take the hyperbolic sine of

        Returns:
            array: hyperbolic sine of ``array``
        """
        ...

    @abstractmethod
    def sqrt(self, x: Tensor, dtype=None) -> Tensor:
        r"""Returns the square root of ``x``.

        Args:
            x (array): array to take the square root of
            dtype (type): ``dtype`` of the output array

        Returns:
            array: square root of ``x``
        """
        ...

    @abstractmethod
    def sum(self, array: Tensor, axes: Sequence[int] = None):
        r"""Returns the sum of array.

        Args:
            array (array): array to take the sum of
            axes (tuple): axes to sum over

        Returns:
            array: sum of array
        """
        ...

    @abstractmethod
    def tensordot(self, a: Tensor, b: Tensor, axes: Sequence[int]) -> Tensor:
        r"""Returns the tensordot product of ``a`` and ``b``.

        Args:
            a (array): first array to take the tensordot product of
            b (array): second array to take the tensordot product of
            axes (tuple): axes to take the tensordot product over

        Returns:
            array: tensordot product of ``a`` and ``b``
        """
        ...

    @abstractmethod
    def tile(self, array: Tensor, repeats: Sequence[int]) -> Tensor:
        r"""Returns the tiled array.

        Args:
            array (array): array to tile
            repeats (tuple): number of times to tile the array along each axis

        Returns:
            array: tiled array
        """
        ...

    @abstractmethod
    def trace(self, array: Tensor) -> Tensor:
        r"""Returns the trace of array.

        Args:
            array (array): array to take the trace of

        Returns:
            array: trace of array
        """
        ...

    @abstractmethod
    def transpose(self, a: Tensor, perm: Sequence[int] = None):
        r"""Returns the transposed arrays.

        Args:
            a (array): array to transpose
            perm (tuple): permutation to apply to the array

        Returns:
            array: transposed array
        """
        ...

    @abstractmethod
    def unique_tensors(lst: List[Tensor]) -> List[Tensor]:
        r"""Returns the tensors in ``lst`` without duplicates and non-tensors.

        Args:
            lst (list): list of tensors to remove duplicates and non-tensors from.

        Returns:
            list: list of tensors without duplicates and non-tensors.
        """
        ...

    @abstractmethod
    def update_tensor(self, tensor: Tensor, indices: Tensor, values: Tensor) -> Tensor:
        r"""Updates a tensor in place with the given values.

        Args:
            tensor (array): tensor to update
            indices (array): indices to update
            values (array): values to update
        """
        ...

    @abstractmethod
    def update_add_tensor(self, tensor: Tensor, indices: Tensor, values: Tensor) -> Tensor:
        r"""Updates a tensor in place by adding the given values.

        Args:
            tensor (array): tensor to update
            indices (array): indices to update
            values (array): values to add
        """
        ...

    @abstractmethod
    def value_and_gradients(
        self, cost_fn: Callable, parameters: Dict[str, List[Trainable]]
    ) -> Tuple[Tensor, Dict[str, List[Tensor]]]:
        r"""Returns the loss and gradients of the given cost function.

        Args:
            cost_fn (callable): cost function to compute the loss and gradients of
            parameters (dict): parameters to compute the loss and gradients of

        Returns:
            tuple: loss and gradients (dict) of the given cost function
        """
        ...

    @abstractmethod
    def zeros(self, shape: Sequence[int], dtype) -> Tensor:
        r"""Returns an array of zeros with the given shape and ``dtype``.

        Args:
            shape (tuple): shape of the array
            dtype (type): dtype of the array

        Returns:
            array: array of zeros
        """
        ...

    @abstractmethod
    def zeros_like(self, array: Tensor) -> Tensor:
        r"""Returns an array of zeros with the same shape and ``dtype`` as ``array``.

        Args:
            array (array): array to take the shape and ``dtype`` of

        Returns:
            array: array of zeros
        """
        ...

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Methods that build on the basic ops and don't need to be overridden in the backend implementation
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    @property
    def euclidean_opt(self):
        if not self._euclidean_opt:
            self._euclidean_opt = self.DefaultEuclideanOptimizer()
        return self._euclidean_opt

    def eigvals(self, tensor: Tensor) -> Tensor:
        r"""Returns the eigenvalues of a matrix."""
        ...

    def sqrtm(self, tensor: Tensor) -> Tensor:
        r"""Returns the matrix square root."""
        ...

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Methods that build on the basic ops and don't need to be overridden in the backend implementation
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def block(self, blocks: List[List[Tensor]], axes=(-2, -1)) -> Tensor:
        r"""Returns a matrix made from the given blocks.

        Args:
            blocks (list): list of lists of compatible blocks
            axes (tuple): axes to stack the blocks along

        Returns:
            array: matrix made of blocks
        """
        rows = [self.concat(row, axis=axes[1]) for row in blocks]
        return self.concat(rows, axis=axes[0])

    def dagger(self, array: Tensor) -> Tensor:
        r"""Returns the adjoint of ``array``.

        Args:
            array (array): array to take the adjoint of

        Returns:
            array: adjoint of ``array``
        """
        return self.conj(self.transpose(array))

    def unitary_to_orthogonal(self, U):
        r"""Unitary to orthogonal mapping.

        Args:
            U (array): unitary matrix in ``U(n)``

        Returns:
            array: orthogonal matrix in :math:`O(2n)`
        """
        X = self.real(U)
        Y = self.imag(U)
        return self.block([[X, -Y], [Y, X]])

    def random_symplectic(self, num_modes: int = 1, max_r: float = 1.0) -> Tensor:
        r"""A random symplectic matrix in ``Sp(2*num_modes)``.

        Squeezing is sampled uniformly from 0.0 to ``max_r`` (1.0 by default).
        """
        if num_modes == 1:
            W = np.exp(1j * np.random.uniform(size=(1, 1)))
            V = np.exp(1j * np.random.uniform(size=(1, 1)))
        else:
            W = unitary_group.rvs(dim=num_modes)
            V = unitary_group.rvs(dim=num_modes)
        r = np.random.uniform(low=0.0, high=max_r, size=num_modes)
        OW = self.unitary_to_orthogonal(W)
        OV = self.unitary_to_orthogonal(V)
        dd = self.diag(self.concat([self.exp(-r), np.exp(r)], axis=0))
        return OW @ dd @ OV

    def random_orthogonal(self, num_modes: int = 1) -> Tensor:
        """A random orthogonal matrix in :math:`O(2*num_modes)`."""
        if num_modes == 1:
            W = self.exp(1j * np.random.uniform(size=(1, 1)))
        else:
            W = unitary_group.rvs(dim=num_modes)
        return self.unitary_to_orthogonal(W)

    def single_mode_to_multimode_vec(self, vec, num_modes: int):
        r"""Apply the same 2-vector (i.e. single-mode) to a larger number of modes."""
        if vec.shape[-1] != 2:
            raise ValueError("vec must be 2-dimensional (i.e. single-mode)")
        x, y = vec[..., -2], vec[..., -1]
        vec = self.concat([self.tile([x], [num_modes]), self.tile([y], [num_modes])], axis=-1)
        return vec

    def single_mode_to_multimode_mat(self, mat: Tensor, num_modes: int):
        r"""Apply the same :math:`2\times 2` matrix (i.e. single-mode) to a larger number of modes."""
        if mat.shape[-2:] != (2, 2):
            raise ValueError("mat must be a single-mode (2x2) matrix")
        mat = self.diag(
            self.tile(self.expand_dims(mat, axis=-1), (1, 1, num_modes))
        )  # shape [2,2,N,N]
        mat = self.reshape(self.transpose(mat, (0, 2, 1, 3)), [2 * num_modes, 2 * num_modes])
        return mat

    @staticmethod
    @lru_cache()
    def Xmat(num_modes: int):
        r"""Returns the matrix :math:`X_n = \begin{bmatrix}0 & I_n\\ I_n & 0\end{bmatrix}.`

        Args:
            num_modes (int): positive integer

        Returns:
            array: :math:`2N\times 2N` array
        """
        I = np.identity(num_modes)
        O = np.zeros((num_modes, num_modes))
        return np.block([[O, I], [I, O]])

    @staticmethod
    @lru_cache()
    def rotmat(num_modes: int):
        "Rotation matrix from quadratures to complex amplitudes."
        I = np.identity(num_modes)
        return np.sqrt(0.5) * np.block([[I, 1j * I], [I, -1j * I]])

    @staticmethod
    @lru_cache()
    def J(num_modes: int):
        """Symplectic form."""
        I = np.identity(num_modes)
        O = np.zeros_like(I)
        return np.block([[O, I], [-I, O]])

    def add_at_modes(
        self, old: Tensor, new: Optional[Tensor], modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        """Adds two phase-space tensors (cov matrices, displacement vectors, etc..) on the specified modes."""
        if new is None:
            return old
        N = old.shape[-1] // 2
        indices = modes + [m + N for m in modes]
        return self.update_add_tensor(
            old, list(product(*[indices] * len(new.shape))), self.reshape(new, -1)
        )

    def left_matmul_at_modes(
        self, a_partial: Tensor, b_full: Tensor, modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        r"""Left matrix multiplication of a partial matrix and a full matrix.

        It assumes that that ``a_partial`` is a matrix operating on M modes and that ``modes`` is a
        list of ``M`` integers, i.e., it will apply ``a_partial`` on the corresponding ``M`` modes
        of ``b_full`` from the left.

        Args:
            a_partial (array): :math:`2M\times 2M` array
            b_full (array): :math:`2N\times 2N` array
            modes (list): list of ``M`` modes to perform the multiplication on

        Returns:
            array: :math:`2N\times 2N` array
        """
        if a_partial is None:
            return b_full
        N = b_full.shape[-1] // 2
        indices = self.astensor(modes + [m + N for m in modes])
        b_rows = self.gather(b_full, indices, axis=0)
        b_rows = self.matmul(a_partial, b_rows)
        return self.update_tensor(b_full, indices[:, None], b_rows)

    def right_matmul_at_modes(
        self, a_full: Tensor, b_partial: Tensor, modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        r"""Right matrix multiplication of a full matrix and a partial matrix.

        It assumes that that ``b_partial`` is a matrix operating on ``M`` modes and that ``modes``
        is a list of ``M`` integers, i.e., it will apply ``b_partial`` on the corresponding M modes
        of ``a_full`` from the right.

        Args:
            a_full (array): :math:`2N\times 2N` array
            b_partial (array): :math:`2M\times 2M` array
            modes (list): list of `M` modes to perform the multiplication on

        Returns:
            array: :math:`2N\times 2N` array
        """
        return self.transpose(
            self.left_matmul_at_modes(self.transpose(b_partial), self.transpose(a_full), modes)
        )

    def matvec_at_modes(
        self, mat: Optional[Tensor], vec: Tensor, modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        """Matrix-vector multiplication between a phase-space matrix and a vector in the specified modes."""
        if mat is None:
            return vec
        N = vec.shape[-1] // 2
        indices = self.astensor(modes + [m + N for m in modes])
        updates = self.matvec(mat, self.gather(vec, indices))
        return self.update_tensor(vec, indices[:, None], updates)

    def all_diagonals(self, rho: Tensor, real: bool) -> Tensor:
        """Returns all the diagonals of a density matrix."""
        cutoffs = rho.shape[: rho.ndim // 2]
        rho = self.reshape(rho, (int(np.prod(cutoffs)), int(np.prod(cutoffs))))
        diag = self.diag_part(rho)
        if real:
            return self.real(self.reshape(diag, cutoffs))
        else:
            return self.reshape(diag, cutoffs)

    def poisson(self, max_k: int, rate: Tensor) -> Tensor:
        """Poisson distribution up to ``max_k``."""
        k = self.arange(max_k)
        rate = self.cast(rate, k.dtype)
        return self.exp(k * self.log(rate + 1e-9) - rate - self.lgamma(k + 1.0))

    def binomial_conditional_prob(self, success_prob: Tensor, dim_out: int, dim_in: int):
        """:math:`P(out|in) = binom(in, out) * (1-success_prob)**(in-out) * success_prob**out`."""
        in_ = self.arange(dim_in)[None, :]
        out_ = self.arange(dim_out)[:, None]
        return (
            self.cast(binom(in_, out_), in_.dtype)
            * self.pow(success_prob, out_)
            * self.pow(1.0 - success_prob, self.maximum(in_ - out_, 0.0))
        )

    def convolve_probs_1d(self, prob: Tensor, other_probs: List[Tensor]) -> Tensor:
        """Convolution of a joint probability with a list of single-index probabilities."""

        if prob.ndim > 3 or len(other_probs) > 3:
            raise ValueError("cannot convolve arrays with more than 3 axes")
        if not all([q.ndim == 1 for q in other_probs]):
            raise ValueError("other_probs must contain 1d arrays")
        if not all([len(q) == s for q, s in zip(other_probs, prob.shape)]):
            raise ValueError("The length of the 1d prob vectors must match shape of prob")

        q = other_probs[0]
        for q_ in other_probs[1:]:
            q = q[..., None] * q_[(None,) * q.ndim + (slice(None),)]

        return self.convolve_probs(prob, q)

    def convolve_probs(self, prob: Tensor, other: Tensor) -> Tensor:
        r"""Convolve two probability distributions (up to 3D) with the same shape.

        Note that the output is not guaranteed to be a complete joint probability,
        as it's computed only up to the dimension of the base probs.
        """

        if prob.ndim > 3 or other.ndim > 3:
            raise ValueError("cannot convolve arrays with more than 3 axes")
        if not prob.shape == other.shape:
            raise ValueError("prob and other must have the same shape")

        prob_padded = self.pad(prob, [(s - 1, 0) for s in other.shape])
        other_reversed = other[(slice(None, None, -1),) * other.ndim]
        return self.convolution(
            prob_padded[None, ..., None],
            other_reversed[..., None, None],
            padding="VALID",  # TODO: do we need to specify this?
            data_format="N"
            + ("HD"[: other.ndim - 1])[::-1]
            + "WC",  # TODO: rewrite this to be more readable (do we need it?)
        )[0, ..., 0]

    def euclidean_to_symplectic(self, S: Matrix, dS_euclidean: Matrix) -> Matrix:
        r"""Convert the Euclidean gradient to a Riemannian gradient on the
        tangent bundle of the symplectic manifold.

        Implemented from:
            Wang J, Sun H, Fiori S. A Riemannian‐steepest‐descent approach
            for optimization on the real symplectic group.
            Mathematical Methods in the Applied Sciences. 2018 Jul 30;41(11):4273-86.

        Args:
            S (Matrix): symplectic matrix
            dS_euclidean (Matrix): Euclidean gradient tensor

        Returns:
            Matrix: symplectic gradient tensor
        """
        Jmat = self.J(S.shape[-1] // 2)
        Z = self.matmul(self.transpose(S), dS_euclidean)
        return 0.5 * (Z + self.matmul(self.matmul(Jmat, self.transpose(Z)), Jmat))



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union, Optional, List, Tuple, Dict
from mrmustard.types import Tensor, Trainable
from mrmustard import settings
from mrmustard.lab.abstract import Transformation
from mrmustard.utils.parametrized import Parametrized
from mrmustard.utils import training
from mrmustard.physics import gaussian

__all__ = [
    "Dgate",
    "Sgate",
    "Rgate",
    "Pgate",
    "Ggate",
    "BSgate",
    "MZgate",
    "S2gate",
    "CZgate",
    "CXgate",
    "Interferometer",
    "Attenuator",
    "Amplifier",
    "AdditiveNoise",
]


class Dgate(Parametrized, Transformation):
    r"""Displacement gate.

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.

    If a parameter is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats. One can optionally set bounds for each
    parameter, which the optimizer will respect.

    Args:
        x (float or List[float]): the list of displacements along the x axis
        x_bounds (float, float): bounds for the displacement along the x axis
        x_trainable (bool): whether x is a trainable variable
        y (float or List[float]): the list of displacements along the y axis
        y_bounds (float, float): bounds for the displacement along the y axis
        y_trainable bool: whether y is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        x: Union[Optional[float], Optional[List[float]]] = 0.0,
        y: Union[Optional[float], Optional[List[float]]] = 0.0,
        x_trainable: bool = False,
        y_trainable: bool = False,
        x_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        y_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            x=x,
            y=y,
            x_trainable=x_trainable,
            y_trainable=y_trainable,
            x_bounds=x_bounds,
            y_bounds=y_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def d_vector(self):
        return gaussian.displacement(self.x, self.y, settings.HBAR)


class Sgate(Parametrized, Transformation):
    r"""Squeezing gate.

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.

    If a parameter is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats. One can optionally set bounds for each
    parameter, which the optimizer will respect.

    Args:
        r (float or List[float]): the list of squeezing magnitudes
        r_bounds (float, float): bounds for the squeezing magnitudes
        r_trainable (bool): whether r is a trainable variable
        phi (float or List[float]): the list of squeezing angles
        phi_bounds (float, float): bounds for the squeezing angles
        phi_trainable bool: whether phi is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        r: Union[Optional[float], Optional[List[float]]] = 0.0,
        phi: Union[Optional[float], Optional[List[float]]] = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            r=r,
            phi=phi,
            r_trainable=r_trainable,
            phi_trainable=phi_trainable,
            r_bounds=r_bounds,
            phi_bounds=phi_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.squeezing_symplectic(self.r, self.phi)


class Rgate(Parametrized, Transformation):
    r"""Rotation gate.

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.

    If a parameter is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats. One can optionally set bounds for each
    parameter, which the optimizer will respect.

    Args:
        modes (List[int]): the list of modes this gate is applied to
        angle (float or List[float]): the list of rotation angles
        angle_bounds (float, float): bounds for the rotation angles
        angle_trainable bool: whether angle is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        angle: Union[Optional[float], Optional[List[float]]] = 0.0,
        angle_trainable: bool = False,
        angle_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            angle=angle,
            angle_trainable=angle_trainable,
            angle_bounds=angle_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.rotation_symplectic(self.angle)


class Pgate(Parametrized, Transformation):
    r"""Quadratic phase gate.

    If len(modes) > 1 the gate is applied in parallel to all of the modes provided. If a parameter
    is a single float, the parallel instances of the gate share that parameter. To apply
    mode-specific values use a list of floats. One can optionally set bounds for each parameter,
    which the optimizer will respect.

    Args:
        modes (List[int]): the list of modes this gate is applied to
        shearing (float or List[float]): the list of shearing parameters
        shearing_bounds (float, float): bounds for the shearing parameters
        shearing_trainable bool: whether shearing is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        shearing: Union[Optional[float], Optional[List[float]]] = 0.0,
        shearing_trainable: bool = False,
        shearing_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            shearing=shearing,
            shearing_trainable=shearing_trainable,
            shearing_bounds=shearing_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.quadratic_phase(self.shearing)


class CXgate(Parametrized, Transformation):
    r"""Controlled X gate.

    It applies to a single pair of modes. One can optionally set bounds for each parameter, which
    the optimizer will respect.

    Args:
        s (float): control parameter
        s_bounds (float, float): bounds for the control angle
        s_trainable (bool): whether s is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        s: Optional[float] = 0.0,
        s_trainable: bool = False,
        s_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            s=s,
            s_trainable=s_trainable,
            s_bounds=s_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.controlled_X(self.s)


class CZgate(Parametrized, Transformation):
    r"""Controlled Z gate.

    It applies to a single pair of modes. One can optionally set bounds for each parameter, which
    the optimizer will respect.

    Args:
        s (float): control parameter
        s_bounds (float, float): bounds for the control angle
        s_trainable (bool): whether s is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        s: Optional[float] = 0.0,
        s_trainable: bool = False,
        s_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            s=s,
            s_trainable=s_trainable,
            s_bounds=s_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.controlled_Z(self.s)


class BSgate(Parametrized, Transformation):
    r"""Beam splitter gate.

    It applies to a single pair of modes.
    One can optionally set bounds for each parameter, which the optimizer will respect.

    Args:
        theta (float): the transmissivity angle
        theta_bounds (float, float): bounds for the transmissivity angle
        theta_trainable (bool): whether theta is a trainable variable
        phi (float): the phase angle
        phi_bounds (float, float): bounds for the phase angle
        phi_trainable bool: whether phi is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        theta: Optional[float] = 0.0,
        phi: Optional[float] = 0.0,
        theta_trainable: bool = False,
        phi_trainable: bool = False,
        theta_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            theta=theta,
            phi=phi,
            theta_trainable=theta_trainable,
            phi_trainable=phi_trainable,
            theta_bounds=theta_bounds,
            phi_bounds=phi_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.beam_splitter_symplectic(self.theta, self.phi)

    def _validate_modes(self, modes):
        if len(modes) != 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be 2). Perhaps you are looking for Interferometer."
            )


class MZgate(Parametrized, Transformation):
    r"""Mach-Zehnder gate.

    It supports two conventions:
        1. if ``internal=True``, both phases act inside the interferometer: ``phi_a`` on the upper arm, ``phi_b`` on the lower arm;
        2. if ``internal = False``, both phases act on the upper arm: ``phi_a`` before the first BS, ``phi_b`` after the first BS.

    One can optionally set bounds for each parameter, which the optimizer will respect.

    Args:
        phi_a (float): the phase in the upper arm of the MZ interferometer
        phi_a_bounds (float, float): bounds for phi_a
        phi_a_trainable (bool): whether phi_a is a trainable variable
        phi_b (float): the phase in the lower arm or external of the MZ interferometer
        phi_b_bounds (float, float): bounds for phi_b
        phi_b_trainable (bool): whether phi_b is a trainable variable
        internal (bool): whether phases are both in the internal arms (default is False)
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        phi_a: Optional[float] = 0.0,
        phi_b: Optional[float] = 0.0,
        phi_a_trainable: bool = False,
        phi_b_trainable: bool = False,
        phi_a_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        phi_b_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        internal: bool = False,
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            phi_a=phi_a,
            phi_b=phi_b,
            phi_a_trainable=phi_a_trainable,
            phi_b_trainable=phi_b_trainable,
            phi_a_bounds=phi_a_bounds,
            phi_b_bounds=phi_b_bounds,
            internal=internal,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.mz_symplectic(self.phi_a, self.phi_b, internal=self._internal)

    def _validate_modes(self, modes):
        if len(modes) != 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be 2). Perhaps you are looking for Interferometer?"
            )


class S2gate(Parametrized, Transformation):
    r"""Two-mode squeezing gate.

    It applies to a single pair of modes. One can optionally set bounds for each parameter, which the optimizer will respect.

    Args:
        r (float): the squeezing magnitude
        r_bounds (float, float): bounds for the squeezing magnitude
        r_trainable (bool): whether r is a trainable variable
        phi (float): the squeezing angle
        phi_bounds (float, float): bounds for the squeezing angle
        phi_trainable bool: whether phi is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        r: Optional[float] = 0.0,
        phi: Optional[float] = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            r=r,
            phi=phi,
            r_trainable=r_trainable,
            phi_trainable=phi_trainable,
            r_bounds=r_bounds,
            phi_bounds=phi_bounds,
            modes=modes,
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.two_mode_squeezing_symplectic(self.r, self.phi)

    def _validate_modes(self, modes):
        if len(modes) != 2:
            raise ValueError(f"Invalid number of modes: {len(modes)} (should be 2")


class Interferometer(Parametrized, Transformation):
    r"""N-mode interferometer.

    It corresponds to a Ggate with zero mean and a ``2N x 2N`` orthogonal symplectic matrix.

    Args:
        orthogonal (2d array): a valid orthogonal matrix. For N modes it must have shape `(2N,2N)`
        orthogonal_trainable (bool): whether orthogonal is a trainable variable
    """

    def __init__(
        self,
        num_modes: int,
        orthogonal: Optional[Tensor] = None,
        orthogonal_trainable: bool = False,
    ):
        if orthogonal is None:
            orthogonal = training.new_orthogonal(num_modes=num_modes)
        super().__init__(
            orthogonal=orthogonal,
            orthogonal_trainable=orthogonal_trainable,
            orthogonal_bounds=(None, None),
            modes=list(range(num_modes)),
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return self.orthogonal

    def _validate_modes(self, modes):
        if len(modes) != self.orthogonal.shape[1] // 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be {self.orthogonal.shape[1] // 2})"
            )

    @property
    def trainable_parameters(self) -> Dict[str, List[Trainable]]:
        return {
            "symplectic": [],
            "orthogonal": [self.orthogonal] if self._orthogonal_trainable else [],
            "euclidean": [],
        }


class Ggate(Parametrized, Transformation):
    r"""A generic N-mode Gaussian unitary transformation with zero displacement.

    If a symplectic matrix is not provided, one will be picked at random with effective squeezing
    strength ``r`` in ``[0, 1]`` for each mode.

    Args:
        num_modes (int): the number of modes this gate is acting on.
        symplectic (2d array): a valid symplectic matrix in XXPP order. For N modes it must have shape ``(2N,2N)``.
        symplectic_trainable (bool): whether symplectic is a trainable variable.
    """

    def __init__(
        self,
        num_modes: int,
        symplectic: Optional[Tensor] = None,
        symplectic_trainable: bool = False,
    ):
        if symplectic is None:
            symplectic = training.new_symplectic(num_modes=num_modes)
        super().__init__(
            symplectic=symplectic,
            symplectic_trainable=symplectic_trainable,
            symplectic_bounds=(None, None),
            modes=list(range(num_modes)),
        )
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return self.symplectic

    def _validate_modes(self, modes):
        if len(modes) != self.symplectic.shape[1] // 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be {self.symplectic.shape[1] // 2})"
            )

    @property
    def trainable_parameters(self) -> Dict[str, List[Trainable]]:
        return {
            "symplectic": [self.symplectic] if self._symplectic_trainable else [],
            "orthogonal": [],
            "euclidean": [],
        }


# ~~~~~~~~~~~~~
# NON-UNITARY
# ~~~~~~~~~~~~~


class Attenuator(Parametrized, Transformation):
    r"""The noisy attenuator channel.

    It corresponds to mixing with a thermal environment and applying the pure loss channel. The pure
    lossy channel is recovered for nbar = 0 (i.e. mixing with vacuum).

    The CPT channel is given by

    .. math::

        X = sqrt(transmissivity) * I
        Y = (1-transmissivity) * (2*nbar + 1) * (hbar / 2) * I

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.
    If ``transmissivity`` is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats.

    One can optionally set bounds for `transmissivity`, which the optimizer will respect.

    Args:
        transmissivity (float or List[float]): the list of transmissivities
        nbar (float): the average number of photons in the thermal state
        transmissivity_trainable (bool): whether transmissivity is a trainable variable
        nbar_trainable (bool): whether nbar is a trainable variable
        transmissivity_bounds (float, float): bounds for the transmissivity
        nbar_bounds (float, float): bounds for the average number of photons in the thermal state
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        transmissivity: Union[Optional[float], Optional[List[float]]] = 1.0,
        nbar: float = 0.0,
        transmissivity_trainable: bool = False,
        nbar_trainable: bool = False,
        transmissivity_bounds: Tuple[Optional[float], Optional[float]] = (0.0, 1.0),
        nbar_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            transmissivity=transmissivity,
            nbar=nbar,
            transmissivity_trainable=transmissivity_trainable,
            nbar_trainable=nbar_trainable,
            transmissivity_bounds=transmissivity_bounds,
            nbar_bounds=nbar_bounds,
            modes=modes,
        )
        self.is_unitary = False
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.loss_XYd(self.transmissivity, self.nbar, settings.HBAR)[0]

    @property
    def Y_matrix(self):
        return gaussian.loss_XYd(self.transmissivity, self.nbar, settings.HBAR)[1]


class Amplifier(Parametrized, Transformation):
    r"""The noisy amplifier channel.

    It corresponds to mixing with a thermal environment and applying a two-mode squeezing gate.

    .. code:: python

        X = sqrt(gain) * I
        Y = (gain-1) * (2*nbar + 1) * (hbar / 2) * I

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.
    If ``gain`` is a single float, the parallel instances of the gate share that parameter.
    To apply mode-specific values use a list of floats.
    One can optionally set bounds for ``gain``, which the optimizer will respect.

    Args:
        gain (float or List[float]): the list of gains (must be > 1)
        nbar (float): the average number of photons in the thermal state
        nbar_trainable (bool): whether nbar is a trainable variable
        gain_trainable (bool): whether gain is a trainable variable
        gain_bounds (float, float): bounds for the gain
        nbar_bounds (float, float): bounds for the average number of photons in the thermal state
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        gain: Union[Optional[float], Optional[List[float]]] = 1.0,
        nbar: float = 0.0,
        gain_trainable: bool = False,
        nbar_trainable: bool = False,
        gain_bounds: Tuple[Optional[float], Optional[float]] = (1.0, None),
        nbar_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            gain=gain,
            gain_trainable=gain_trainable,
            gain_bounds=gain_bounds,
            nbar=nbar,
            nbar_trainable=nbar_trainable,
            nbar_bounds=nbar_bounds,
            modes=modes,
        )
        self.is_unitary = False
        self.is_gaussian = True

    @property
    def X_matrix(self):
        return gaussian.amp_XYd(self.gain, self.nbar, settings.HBAR)[0]

    @property
    def Y_matrix(self):
        return gaussian.amp_XYd(self.gain, self.nbar, settings.HBAR)[1]


class AdditiveNoise(Parametrized, Transformation):
    r"""The additive noise channel.

    Equivalent to an amplifier followed by an attenuator. E.g.,

    .. code-block::

        na,nb = np.random.uniform(size=2)
        tr = np.random.uniform()
        Amplifier(1/tr, nb) >> Attenuator(tr, na) == AdditiveNoise(2*(1-tr)*(1+na+nb)) # evaluates to True

    or equivalent to an attenuator followed by an amplifier:

    .. code-block::

        na,nb = np.random.uniform(size=2)
        amp = 1.0 + np.random.uniform()
        Attenuator(1/amp, nb) >> Amplifier(amp, na) == AdditiveNoise(2*(amp-1)*(1+na+nb))

    Args:
        noise (float or List[float]): the added noise in units of hbar/2
        noise_trainable (bool): whether noise is a trainable variable
        noise_bounds (float, float): bounds for the noise
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    def __init__(
        self,
        noise: Union[Optional[float], Optional[List[float]]] = 0.0,
        noise_trainable: bool = False,
        noise_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            noise=noise, noise_trainable=noise_trainable, noise_bounds=noise_bounds, modes=modes
        )
        self.is_unitary = False
        self.is_gaussian = True

    @property
    def Y_matrix(self):
        return gaussian.noise_Y(self.noise, settings.HBAR)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import warnings
import numpy as np
from rich.table import Table
from rich import print as rprint

from mrmustard.types import *
from mrmustard.utils import graphics
from mrmustard import settings
from mrmustard.utils.xptensor import XPMatrix, XPVector
from mrmustard.physics import gaussian, fock
from mrmustard.math import Math

math = Math()


class State:
    r"""Base class for quantum states."""

    def __init__(
        self,
        cov: Matrix = None,
        means: Vector = None,
        eigenvalues: Array = None,
        symplectic: Matrix = None,
        ket: Array = None,
        dm: Array = None,
        modes: Sequence[int] = None,
        cutoffs: Sequence[int] = None,
        _norm: float = 1.0,
    ):
        r"""Initializes the state.

        Supply either:
            * a covariance matrix and means vector
            * an eigenvalues array and symplectic matrix
            * a fock representation (ket or dm)

        Args:
            cov (Matrix): the covariance matrix
            means (Vector): the means vector
            eigenvalues (Array): the eigenvalues of the covariance matrix
            symplectic (Matrix): the symplectic matrix mapping the thermal state with given eigenvalues to this state
            fock (Array): the Fock representation
            modes (optional, Sequence[int]): the modes in which the state is defined
            cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
            _norm (float, default=1.0): the norm of the state. Warning: only set if you know what you are doing.

        """
        self._purity = None
        self._fock_probabilities = None
        self._cutoffs = cutoffs
        self._cov = cov
        self._means = means
        self._eigenvalues = eigenvalues
        self._symplectic = symplectic
        self._ket = ket
        self._dm = dm
        self._norm = _norm
        if cov is not None and means is not None:
            self.is_gaussian = True
            self.num_modes = cov.shape[-1] // 2
        elif eigenvalues is not None and symplectic is not None:
            self.is_gaussian = True
            self.num_modes = symplectic.shape[-1] // 2
        elif ket is not None or dm is not None:
            self.is_gaussian = False
            self.num_modes = len(ket.shape) if ket is not None else len(dm.shape) // 2
            self._purity = 1.0 if ket is not None else None
        else:
            raise ValueError(
                "State must be initialized with either a covariance matrix and means vector, an eigenvalues array and symplectic matrix, or a fock representation"
            )
        self._modes = modes
        if modes is not None:
            assert (
                len(modes) == self.num_modes
            ), f"Number of modes supplied ({len(modes)}) must match the representation dimension {self.num_modes}"

    @property
    def modes(self):
        r"""Returns the modes of the state."""
        if self._modes is None:
            return list(range(self.num_modes))
        return self._modes

    def indices(self, modes) -> Union[Tuple[int], int]:
        r"""Returns the indices of the given modes.

        Args:
            modes (Sequence[int] or int): the modes or mode

        Returns:
            Tuple[int] or int: a tuple of indices of the given modes or the single index of a single mode
        """
        if isinstance(modes, int):
            return self.modes.index(modes)
        return tuple(self.modes.index(m) for m in modes)

    @property
    def purity(self) -> float:
        if self._purity is None:
            if self.is_gaussian:
                self._purity = gaussian.purity(self.cov, settings.HBAR)
                # TODO: add symplectic representation
            else:
                self._purity = fock.purity(self.fock)  # has to be dm
        return self._purity

    @property
    def is_mixed(self):
        r"""Returns whether the state is mixed."""
        return not self.is_pure

    @property
    def is_pure(self):
        r"""Returns ``True`` if the state is pure and ``False`` otherwise."""
        return np.isclose(self.purity, 1.0, atol=1e-6)

    @property
    def means(self) -> Optional[Vector]:
        r"""Returns the means vector of the state."""
        return self._means

    @property
    def cov(self) -> Optional[Matrix]:
        r"""Returns the covariance matrix of the state."""
        return self._cov

    @property
    def number_stdev(self) -> Vector:
        r"""Returns the square root of the photon number variances (standard deviation) in each mode."""
        if self.is_gaussian:
            return math.sqrt(math.diag_part(self.number_cov))
        else:
            return math.sqrt(
                fock.number_variances(self.fock, is_dm=len(self.fock.shape) == self.num_modes * 2)
            )

    @property
    def cutoffs(self) -> List[int]:
        r"""Returns the cutoff dimensions for each mode."""
        if self._cutoffs is not None:
            return self._cutoffs  # TODO: allow self._cutoffs = [N, None]
        if self._ket is None and self._dm is None:
            return fock.autocutoffs(
                self.number_stdev, self.number_means
            )  # TODO: move autocutoffs in gaussian.py and pass cov, means
        else:  # NOTE: triggered only if the fock representation already exists
            return list(self.fock.shape[: self.num_modes])

    @property
    def shape(self) -> List[int]:
        r"""Returns the shape of the state, accounting for ket/dm representation.

        If the state is in Gaussian representation, the shape is inferred from
        the first two moments of the number operator.
        """
        # NOTE: if we initialize State(dm=pure_dm), self.fock returns the dm, which does not have shape self.cutoffs
        return self.cutoffs if self.is_pure else self.cutoffs + self.cutoffs

    @property
    def fock(self) -> Array:
        r"""Returns the Fock representation of the state."""
        if self._dm is None and self._ket is None:
            _fock = fock.fock_representation(
                self.cov, self.means, shape=self.shape, return_dm=self.is_mixed
            )
            if self.is_mixed:
                self._dm = _fock
                self._ket = None
            else:
                self._ket = _fock
                self._dm = None
        return self._ket if self._ket is not None else self._dm

    @property
    def number_means(self) -> Vector:
        r"""Returns the mean photon number for each mode."""
        if self.is_gaussian:
            return gaussian.number_means(self.cov, self.means, settings.HBAR)
        else:
            return fock.number_means(tensor=self.fock, is_dm=self.is_mixed)

    @property
    def number_cov(self) -> Matrix:
        r"""Returns the complete photon number covariance matrix."""
        if self.is_gaussian:
            return gaussian.number_cov(self.cov, self.means, settings.HBAR)
        else:
            raise NotImplementedError("number_cov not yet implemented for non-gaussian states")

    @property
    def norm(self) -> float:
        r"""Returns the norm of the state."""
        if self.is_gaussian:
            return self._norm
        else:
            return fock.norm(self.fock, self.is_mixed)

    def ket(self, cutoffs: List[int] = None) -> Optional[Tensor]:
        r"""Returns the ket of the state in Fock representation or ``None`` if the state is mixed.

        Args:
            cutoffs List[int or None]: The cutoff dimensions for each mode. If a mode cutoff is
                ``None``, it's guessed automatically.

        Returns:
            Tensor: the ket
        """
        if self.is_mixed:
            return None
        cutoffs = (
            self.cutoffs
            if cutoffs is None
            else [c if c is not None else self.cutoffs[i] for i, c in enumerate(cutoffs)]
        )
        if self.is_gaussian:
            self._ket = fock.fock_representation(
                self.cov, self.means, shape=cutoffs, return_dm=False
            )
        else:  # only fock representation is available
            if self._ket is None:
                return None
            current_cutoffs = [s for s in self._ket.shape[: self.num_modes]]
            if cutoffs != current_cutoffs:
                paddings = [(0, max(0, new - old)) for new, old in zip(cutoffs, current_cutoffs)]
                if any(p != (0, 0) for p in paddings):
                    padded = fock.math.pad(self._ket, paddings, mode="constant")
                else:
                    padded = self._ket
                return padded[tuple([slice(s) for s in cutoffs])]
        return self._ket

    def dm(self, cutoffs: List[int] = None) -> Tensor:
        r"""Returns the density matrix of the state in Fock representation.

        Args:
            cutoffs List[int]: The cutoff dimensions for each mode. If a mode cutoff is ``None``,
                it's automatically computed.

        Returns:
            Tensor: the density matrix
        """
        if cutoffs is None:
            cutoffs = self.cutoffs
        else:
            cutoffs = [c if c is not None else self.cutoffs[i] for i, c in enumerate(cutoffs)]
        if self.is_pure:
            ket = self.ket(cutoffs=cutoffs)
            if ket is not None:
                return fock.ket_to_dm(ket)
        else:
            if self.is_gaussian:
                self._dm = fock.fock_representation(
                    self.cov, self.means, shape=cutoffs * 2, return_dm=True
                )
            elif cutoffs != (current_cutoffs := [s for s in self._dm.shape[: self.num_modes]]):
                paddings = [(0, max(0, new - old)) for new, old in zip(cutoffs, current_cutoffs)]
                if any(p != (0, 0) for p in paddings):
                    padded = fock.math.pad(self._dm, paddings + paddings, mode="constant")
                else:
                    padded = self._dm
                return padded[tuple([slice(s) for s in cutoffs + cutoffs])]
        return self._dm

    def fock_probabilities(self, cutoffs: Sequence[int]) -> Tensor:
        r"""Returns the probabilities in Fock representation.

        If the state is pure, they are the absolute value squared of the ket amplitudes.
        If the state is mixed they are the multi-dimensional diagonals of the density matrix.

        Args:
            cutoffs List[int]: the cutoff dimensions for each mode

        Returns:
            Array: the probabilities
        """
        if self._fock_probabilities is None:
            if self.is_mixed:
                dm = self.dm(cutoffs=cutoffs)
                self._fock_probabilities = fock.dm_to_probs(dm)
            else:
                ket = self.ket(cutoffs=cutoffs)
                self._fock_probabilities = fock.ket_to_probs(ket)
        return self._fock_probabilities

    def primal(self, other: Union[State, Transformation]) -> State:
        r"""Returns the post-measurement state after ``other`` is projected onto ``self``.

        ``other << self`` is other projected onto ``self``.

        If ``other`` is a ``Transformation``, it returns the dual of the transformation applied to
        ``self``: ``other << self`` is like ``self >> other^dual``.

        Note that the returned state is not normalized. To normalize a state you can use
        ``mrmustard.physics.normalize``.
        """
        if issubclass(other.__class__, State):
            remaining_modes = [m for m in other.modes if m not in self.modes]

            if self.is_gaussian and other.is_gaussian:
                prob, cov, means = gaussian.general_dyne(
                    other.cov,
                    other.means,
                    self.cov,
                    self.means,
                    other.indices(self.modes),
                    settings.HBAR,
                )
                if len(remaining_modes) > 0:
                    return State(
                        means=means,
                        cov=cov,
                        modes=remaining_modes,
                        _norm=prob if not getattr(self, "_normalize", False) else 1.0,
                    )
                else:
                    return prob
            else:  # either self or other is not gaussian
                other_cutoffs = [
                    None if m not in self.modes else other.cutoffs[other.indices(m)]
                    for m in other.modes
                ]
                try:
                    out_fock = self._preferred_projection(other, other.indices(self.modes))
                except AttributeError:
                    # matching other's cutoffs
                    self_cutoffs = [other.cutoffs[other.indices(m)] for m in self.modes]
                    out_fock = fock.contract_states(
                        stateA=other.ket(other_cutoffs)
                        if other.is_pure
                        else other.dm(other_cutoffs),
                        stateB=self.ket(self_cutoffs) if self.is_pure else self.dm(self_cutoffs),
                        a_is_mixed=other.is_mixed,
                        b_is_mixed=self.is_mixed,
                        modes=other.indices(self.modes),  # TODO: change arg name to indices
                        normalize=self._normalize if hasattr(self, "_normalize") else False,
                    )
                if len(remaining_modes) > 0:
                    return (
                        State(dm=out_fock, modes=remaining_modes)
                        if other.is_mixed or self.is_mixed
                        else State(ket=out_fock, modes=remaining_modes)
                    )
                else:
                    return (
                        fock.math.abs(out_fock) ** 2
                        if other.is_pure and self.is_pure
                        else fock.math.abs(out_fock)
                    )
        else:
            try:
                return other.dual(self)
            except AttributeError:
                raise TypeError(
                    f"Cannot apply {other.__class__.__qualname__} to {self.__class__.__qualname__}"
                )

    def __and__(self, other: State) -> State:
        r"""Concatenates two states."""
        if self.is_gaussian and other.is_gaussian:
            cov = gaussian.join_covs([self.cov, other.cov])
            means = gaussian.join_means([self.means, other.means])
            return State(
                cov=cov, means=means, modes=self.modes + [m + self.num_modes for m in other.modes]
            )
        else:
            raise NotImplementedError(
                "Concatenation of non-gaussian states is not implemented yet."
            )

    def __getitem__(self, item):
        "setting the modes of a state (same API of `Transformation`)"
        if isinstance(item, int):
            item = [item]
        elif isinstance(item, Iterable):
            item = list(item)
        else:
            raise TypeError("item must be int or iterable")
        if len(item) != self.num_modes:
            raise ValueError(
                f"there are {self.num_modes} modes (item has {len(item)} elements, perhaps you're looking for .get_modes()?)"
            )
        self._modes = item
        return self

    def get_modes(self, item):
        r"""Returns the state on the given modes."""
        if isinstance(item, int):
            item = [item]
        elif isinstance(item, Iterable):
            item = list(item)
        else:
            raise TypeError("item must be int or iterable")
        if self.is_gaussian:
            cov, _, _ = gaussian.partition_cov(self.cov, item)
            means, _ = gaussian.partition_means(self.means, item)
            return State(cov=cov, means=means, modes=item)
        else:
            fock_partitioned = fock.trace(
                self.dm(self.cutoffs), [m for m in range(self.num_modes) if m not in item]
            )
            return State(dm=fock_partitioned, modes=item)

    def __eq__(self, other):
        r"""Returns whether the states are equal."""
        if self.num_modes != other.num_modes:
            return False
        if not np.isclose(self.purity, other.purity, atol=1e-6):
            return False
        if self.is_gaussian and other.is_gaussian:
            if not np.allclose(self.means, other.means, atol=1e-6):
                return False
            if not np.allclose(self.cov, other.cov, atol=1e-6):
                return False
            return True
        if self.is_pure and other.is_pure:
            return np.allclose(
                self.ket(cutoffs=other.cutoffs), other.ket(cutoffs=other.cutoffs), atol=1e-6
            )
        else:
            return np.allclose(
                self.dm(cutoffs=other.cutoffs), other.dm(cutoffs=other.cutoffs), atol=1e-6
            )

    def __rshift__(self, other):
        r"""Applies other (a Transformation) to self (a State), e.g., ``Coherent(x=0.1) >> Sgate(r=0.1)``."""
        if issubclass(other.__class__, State):
            raise TypeError(
                f"Cannot apply {other.__class__.__qualname__} to a state. Are you looking for the << operator?"
            )
        return other.primal(self)

    def __lshift__(self, other: State):
        r"""Implements projection onto a state or the dual transformation applied on a state.

        E.g., ``self << other`` where other is a ``State`` and ``self`` is either a ``State`` or a ``Transformation``.
        """
        return other.primal(self)

    def __add__(self, other: State):
        r"""Implements a mixture of states (only available in fock representation for the moment)."""
        if not isinstance(other, State):
            raise TypeError(f"Cannot add {other.__class__.__qualname__} to a state")
        warnings.warn("mixing states forces conversion to fock representation", UserWarning)
        return State(dm=self.dm(self.cutoffs) + other.dm(self.cutoffs))

    def __rmul__(self, other):
        r"""Implements multiplication by a scalar from the left.

        E.g., ``0.5 * psi``.
        """
        if self.is_gaussian:
            warnings.warn(
                "scalar multiplication forces conversion to fock representation", UserWarning
            )
            self.fock  # trigger creation of fock representation
        if self._dm is not None:
            return State(dm=self.dm() * other, modes=self.modes)
        elif self._ket is not None:
            return State(ket=self.ket() * other, modes=self.modes)
        else:
            raise ValueError("No fock representation available")

    def __truediv__(self, other):
        r"""Implements division by a scalar from the left.

        E.g. ``psi / 0.5``
        """
        if self.is_gaussian:
            warnings.warn("scalar division forces conversion to fock representation", UserWarning)
            self.fock
        if self._dm is not None:
            return State(dm=self.dm() / other, modes=self.modes)
        elif self._ket is not None:
            return State(ket=self.ket() / other, modes=self.modes)
        else:
            raise ValueError("No fock representation available")

    def _repr_markdown_(self):
        table = (
            f"#### {self.__class__.__qualname__}\n\n"
            + "| Purity | Norm | Num modes | Bosonic size | Gaussian | Fock |\n"
            + "| :----: | :----: | :----: | :----: | :----: | :----: |\n"
            + f"| {self.purity :.2e} | {self.norm :.2e} | {self.num_modes} | {'1' if self.is_gaussian else 'N/A'} | {'✅' if self.is_gaussian else '❌'} | {'✅' if self._ket is not None or self._dm is not None else '❌'} |"
        )

        if self.num_modes == 1:
            graphics.mikkel_plot(math.asnumpy(self.dm(cutoffs=self.cutoffs)))

        if settings.DEBUG:
            detailed_info = f"\ncov={repr(self.cov)}\n" + f"means={repr(self.means)}\n"
            return f"{table}\n{detailed_info}"

        return table



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

__all__ = ["Circuit"]


from mrmustard.types import Matrix, Vector
from typing import List, Tuple
from mrmustard.utils.parametrized import Parametrized
from mrmustard.utils.xptensor import XPMatrix, XPVector
from mrmustard.lab.abstract import Transformation
from mrmustard.lab.abstract import State


class Circuit(Transformation, Parametrized):
    def __init__(self, ops: List = []):
        self._ops: List = [o for o in ops]
        self.reset()

    def reset(self):
        self._compiled: bool = False
        self._modes: List[int] = []

    @property
    def num_modes(self) -> int:  # TODO: improve this
        all_modes = set()
        for op in self._ops:
            all_modes = all_modes | set(op.modes)
        return len(all_modes)

    def primal(self, state: State) -> State:
        for op in self._ops:
            state = op.primal(state)
        return state

    def dual(self, state: State) -> State:
        for op in reversed(self._ops):
            state = op.dual(state)
        return state

    @property
    def XYd(
        self,
    ) -> Tuple[Matrix, Matrix, Vector]:  # NOTE: Overriding Transformation.XYd for efficiency
        X = XPMatrix(like_1=True)
        Y = XPMatrix(like_0=True)
        d = XPVector()
        for op in self._ops:
            opx, opy, opd = op.XYd
            opX = XPMatrix.from_xxpp(opx, modes=(op.modes, op.modes), like_1=True)
            opY = XPMatrix.from_xxpp(opy, modes=(op.modes, op.modes), like_0=True)
            opd = XPVector.from_xxpp(opd, modes=op.modes)
            if opX.shape is not None and opX.shape[-1] == 1 and len(op.modes) > 1:
                opX = opX.clone(len(op.modes), modes=(op.modes, op.modes))
            if opY.shape is not None and opY.shape[-1] == 1 and len(op.modes) > 1:
                opY = opY.clone(len(op.modes), modes=(op.modes, op.modes))
            if opd.shape is not None and opd.shape[-1] == 1 and len(op.modes) > 1:
                opd = opd.clone(len(op.modes), modes=op.modes)
            X = opX @ X
            Y = opX @ Y @ opX.T + opY
            d = opX @ d + opd
        return X.to_xxpp(), Y.to_xxpp(), d.to_xxpp()

    @property
    def is_gaussian(self):
        return all(op.is_gaussian for op in self._ops)

    def extend(self, obj):  # TODO: remove
        result = self._ops.extend(obj)
        self.reset()

    def append(self, obj):  # TODO: remove
        result = self._ops.append(obj)
        self.reset()

    def __len__(self):
        return len(self._ops)

    def _repr_markdown_(self) -> str:
        """Markdown string to display the object on ipython notebooks."""
        return f"Circuit | {len(self._ops)} ops | compiled = `{self._compiled}`"

    def __repr__(self) -> str:
        ops_repr = [repr(op) for op in self._ops]
        return "Circuit([" + ",".join(ops_repr) + "])"



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The ``math`` module contains low-level functions for performing mathematical operations.

It is recommended that users access the backends using the an instance of the :class:`Math` class rather than the backends themselves.

The Math class is a wrapper that passes the calls to the currently active backend, which is determined by
the ``BACKEND`` parameter in ``mrmustard.settings`` (the default is ``tensorflow``).

The advantage of using the Math class is that the same code can run on different backends, allowing for a
greater degree of flexibility and code reuse.

.. code-block::

    from mrmustard.math import Math
    math = Math()
    math.cos(x)  # tensorflow backend

    from mrmustard import settings
    settings.BACKEND = 'torch'

    math.cos(x)  # torch backend
"""


from mrmustard import settings
import importlib

if importlib.util.find_spec("tensorflow"):
    from mrmustard.math.tensorflow import TFMath
if importlib.util.find_spec("torch"):
    from mrmustard.math.torch import TorchMath


class Math:
    r"""
    This class is a switcher for performing math operations on the currently active backend.
    """

    def __getattribute__(self, name):
        if settings.backend == "tensorflow":
            return object.__getattribute__(TFMath(), name)
        elif settings.backend == "torch":
            return object.__getattribute__(TorchMath(), name)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rich.progress import Progress, TextColumn, BarColumn, TimeRemainingColumn
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib import cm
import numpy as np
from mrmustard.types import *
from mrmustard import settings
from numba import njit
from copy import copy


class Progressbar:
    "A spiffy loading bar to display the progress during an optimization."

    def __init__(self, max_steps: int):
        self.taskID = None
        if max_steps == 0:
            self.bar = Progress(
                TextColumn("Step {task.completed}/∞"),
                BarColumn(),
                TextColumn("Cost = {task.fields[loss]:.5f}"),
            )
        else:
            self.bar = Progress(
                TextColumn("Step {task.completed}/{task.total} | {task.fields[speed]:.1f} it/s"),
                BarColumn(),
                TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                TextColumn("Cost = {task.fields[loss]:.5f} | ⏳ "),
                TimeRemainingColumn(),
            )
        self.taskID = self.bar.add_task(
            description="Optimizing...",
            start=max_steps > 0,
            speed=0.0,
            total=max_steps,
            loss=1.0,
            refresh=True,
        )

    def step(self, loss):
        speed = self.bar.tasks[0].speed or 0.0
        self.bar.update(self.taskID, advance=1, refresh=True, speed=speed, loss=loss)

    def __enter__(self):
        return self.bar.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.bar.__exit__(exc_type, exc_val, exc_tb)


def plot_wigner(rho, xvec, pvec, hbar):
    r"""Calculates the discretized Wigner function of the specified mode.

    Lifted from `strawberryfields <https://github.com/XanaduAI/strawberryfields/blob/master/strawberryfields/backends/states.py#L725>`_

    Args:
        rho (complex array): the state in Fock representation (can be pure or mixed)
        xvec (array): array of discretized :math:`x` quadrature values
        pvec (array): array of discretized :math:`p` quadrature values
        hbar (float): the value of ``\hbar``
    """

    Q, P = np.meshgrid(xvec, pvec)
    cutoff = rho.shape[-1]
    A = (Q + P * 1.0j) / (2 * np.sqrt(hbar / 2))

    Wlist = np.array([np.zeros(np.shape(A), dtype=complex) for k in range(cutoff)])

    # Wigner function for |0><0|
    Wlist[0] = np.exp(-2.0 * np.abs(A) ** 2) / np.pi

    # W = rho(0,0)W(|0><0|)
    W = np.real(rho[0, 0]) * np.real(Wlist[0])

    for n in range(1, cutoff):
        Wlist[n] = (2.0 * A * Wlist[n - 1]) / np.sqrt(n)
        W += 2 * np.real(rho[0, n] * Wlist[n])

    for m in range(1, cutoff):
        temp = copy(Wlist[m])
        # Wlist[m] = Wigner function for |m><m|
        Wlist[m] = (2 * np.conj(A) * temp - np.sqrt(m) * Wlist[m - 1]) / np.sqrt(m)

        # W += rho(m,m)W(|m><m|)
        W += np.real(rho[m, m] * Wlist[m])

        for n in range(m + 1, cutoff):
            temp2 = (2 * A * Wlist[n - 1] - np.sqrt(m) * temp) / np.sqrt(n)
            temp = copy(Wlist[n])
            # Wlist[n] = Wigner function for |m><n|
            Wlist[n] = temp2

            # W += rho(m,n)W(|m><n|) + rho(n,m)W(|n><m|)
            W += 2 * np.real(rho[m, n] * Wlist[n])

    return W / hbar


def mikkel_plot(rho: np.ndarray, filename: str = "", xbounds=(-6, 6), ybounds=(-6, 6)):
    X = np.linspace(xbounds[0], xbounds[1], 200)
    P = np.linspace(ybounds[0], ybounds[1], 200)
    W = plot_wigner(rho, X, P, settings.HBAR)
    ProbX = np.sum(W, axis=0)
    ProbP = np.sum(W, axis=1)

    ### PLOTTING ###

    fig, ax = plt.subplots(
        2, 2, figsize=(6, 6), gridspec_kw={"width_ratios": [2, 1], "height_ratios": [1, 2]}
    )
    ticks = [-5, 0, 5]
    xlim = [X[0], X[-1]]
    plim = [P[0], P[-1]]
    grid = False
    plt.subplots_adjust(wspace=0.05, hspace=0.05)

    # Wigner function
    ax[1][0].contourf(X, P, W, 60, cmap=cm.RdBu, vmin=-abs(W).max(), vmax=abs(W).max())
    ax[1][0].set_xlabel("$x$", fontsize=12)
    ax[1][0].set_ylabel("$p$", fontsize=12)
    ax[1][0].get_xaxis().set_ticks(ticks)
    ax[1][0].get_yaxis().set_ticks(ticks)
    ax[1][0].tick_params(direction="in")
    ax[1][0].set_xlim(xlim)
    ax[1][0].set_ylim(plim)
    ax[1][0].grid(grid)

    # X quadrature probability distribution
    ax[0][0].fill(X, ProbX, color=cm.RdBu(0.5))
    ax[0][0].plot(X, ProbX)
    ax[0][0].get_xaxis().set_ticks(ticks)
    ax[0][0].xaxis.set_ticklabels([])
    ax[0][0].get_yaxis().set_ticks([])
    ax[0][0].tick_params(direction="in")
    ax[0][0].set_ylabel("Prob($x$)", fontsize=12)
    ax[0][0].set_xlim(xlim)
    ax[0][0].set_ylim([0, 1.1 * max(ProbX)])
    ax[0][0].grid(grid)

    # P quadrature probability distribution
    ax[1][1].fill(ProbP, P, color=cm.RdBu(0.5))
    ax[1][1].plot(ProbP, P)
    ax[1][1].get_xaxis().set_ticks([])
    ax[1][1].get_yaxis().set_ticks(ticks)
    ax[1][1].yaxis.set_ticklabels([])
    ax[1][1].tick_params(direction="in")
    ax[1][1].set_xlabel("Prob($p$)", fontsize=12)
    ax[1][1].set_xlim([0, 1.1 * max(ProbP)])
    ax[1][1].set_ylim(plim)
    ax[1][1].grid(grid)

    # Density matrix
    ax[0][1].matshow(abs(rho), cmap=cm.RdBu, vmin=-abs(rho).max(), vmax=abs(rho).max())
    ax[0][1].set_title(r"abs($\rho$)", fontsize=12)
    ax[0][1].tick_params(direction="in")
    ax[0][1].get_xaxis().set_ticks([])
    ax[0][1].get_yaxis().set_ticks([])
    ax[0][1].set_aspect("auto")  # if not set to auto, pytho
    ax[0][1].set_ylabel(f"cutoff = {len(rho)}", fontsize=12)
    ax[0][1].yaxis.set_label_position("right")



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from thewalrus.quantum import is_pure_cov

from mrmustard.types import Matrix, Vector, Scalar
from typing import Tuple, Union, Sequence, List, Any
from mrmustard.utils.xptensor import XPMatrix, XPVector
from mrmustard import settings
from numpy import pi
from mrmustard.math import Math

math = Math()


#  ~~~~~~
#  States
#  ~~~~~~


def vacuum_cov(num_modes: int, hbar: float) -> Matrix:
    r"""Returns the real covariance matrix of the vacuum state.

    Args:
        num_modes (int): number of modes
        hbar (float): value of ``hbar``

    Returns:
        Matrix: vacuum covariance matrix
    """
    return math.eye(num_modes * 2, dtype=math.float64) * hbar / 2


def vacuum_means(num_modes: int, hbar: float) -> Tuple[Matrix, Vector]:
    r"""Returns the real covariance matrix and real means vector of the vacuum state.

    Args:
        num_modes (int): number of modes
        hbar (float): value of ``hbar``

    Returns:
        Matrix, Vector: thermal state covariance matrix or means vector
    """
    return displacement(
        math.zeros(num_modes, dtype="float64"), math.zeros(num_modes, dtype="float64"), hbar
    )


def squeezed_vacuum_cov(r: Vector, phi: Vector, hbar: float) -> Matrix:
    r"""Returns the real covariance matrix and real means vector of a squeezed vacuum state.

    The dimension depends on the dimensions of ``r`` and ``phi``.

    Args:
        r (vector): squeezing magnitude
        phi (vector): squeezing angle
        hbar: value of ``hbar``

    Returns:
        Matrix, Vector: thermal state covariance matrix or means vector
    """
    S = squeezing_symplectic(r, phi)
    return math.matmul(S, math.transpose(S)) * hbar / 2


def thermal_cov(nbar: Vector, hbar: float) -> Tuple[Matrix, Vector]:
    r"""Returns the real covariance matrix and real means vector of a thermal state.

    The dimension depends on the dimensions of ``nbar``.

    Args:
        nbar (vector): average number of photons per mode
        hbar: value of ``hbar``

    Returns:
        Matrix, Vector: thermal state covariance matrix or means vector
    """
    g = (2 * math.atleast_1d(nbar) + 1) * hbar / 2
    return math.diag(math.concat([g, g], axis=-1))


def two_mode_squeezed_vacuum_cov(r: Vector, phi: Vector, hbar: float) -> Matrix:
    r"""Returns the real covariance matrix and real means vector of a two-mode squeezed vacuum state.

    The dimension depends on the dimensions of ``r`` and ``phi``.

    Args:
        r (vector): squeezing magnitude
        phi (vector): squeezing angle
        hbar: value of hbar

    Returns:
        Matrix: two-mode squeezed state covariance matrix
        Vector: two-mode squeezed state means vector
    """
    S = two_mode_squeezing_symplectic(r, phi)
    return math.matmul(S, math.transpose(S)) * hbar / 2


def gaussian_cov(symplectic: Matrix, eigenvalues: Vector = None, hbar: float = 2.0) -> Matrix:
    r"""Returns the covariance matrix of a Gaussian state.

    Args:
        symplectic (Tensor): symplectic matrix of a channel
        eigenvalues (vector): symplectic eigenvalues
        hbar (float): value of hbar

    Returns:
        Tensor: covariance matrix of the Gaussian state
    """
    if eigenvalues is None:
        return math.matmul(symplectic, math.transpose(symplectic))
    else:
        return math.matmul(
            math.matmul(symplectic, math.diag(math.concat([eigenvalues, eigenvalues], axis=0))),
            math.transpose(symplectic),
        )


# ~~~~~~~~~~~~~~~~~~~~~~~~
#  Unitary transformations
# ~~~~~~~~~~~~~~~~~~~~~~~~


def rotation_symplectic(angle: Union[Scalar, Vector]) -> Matrix:
    r"""Symplectic matrix of a rotation gate.

    The dimension depends on the dimension of the angle.

    Args:
        angle (scalar or vector): rotation angles

    Returns:
        Tensor: symplectic matrix of a rotation gate
    """
    angle = math.atleast_1d(angle)
    num_modes = angle.shape[-1]
    x = math.cos(angle)
    y = math.sin(angle)
    return (
        math.diag(math.concat([x, x], axis=0))
        + math.diag(-y, k=num_modes)
        + math.diag(y, k=-num_modes)
    )


def squeezing_symplectic(r: Union[Scalar, Vector], phi: Union[Scalar, Vector]) -> Matrix:
    r"""Symplectic matrix of a squeezing gate.

    The dimension depends on the dimension of ``r`` and ``phi``.

    Args:
        r (scalar or vector): squeezing magnitude
        phi (scalar or vector): rotation parameter

    Returns:
        Tensor: symplectic matrix of a squeezing gate
    """
    r = math.atleast_1d(r)
    phi = math.atleast_1d(phi)
    if r.shape[-1] == 1:
        r = math.tile(r, phi.shape)
    if phi.shape[-1] == 1:
        phi = math.tile(phi, r.shape)
    num_modes = phi.shape[-1]
    cp = math.cos(phi)
    sp = math.sin(phi)
    ch = math.cosh(r)
    sh = math.sinh(r)
    cpsh = cp * sh
    spsh = sp * sh
    return (
        math.diag(math.concat([ch - cpsh, ch + cpsh], axis=0))
        + math.diag(-spsh, k=num_modes)
        + math.diag(-spsh, k=-num_modes)
    )


def displacement(x: Union[Scalar, Vector], y: Union[Scalar, Vector], hbar: float) -> Vector:
    r"""Returns the displacement vector for a displacement by :math:`alpha = x + iy`.
    The dimension depends on the dimensions of ``x`` and ``y``.

    Args:
        x (scalar or vector): real part of displacement
        y (scalar or vector): imaginary part of displacement
        hbar: value of hbar

    Returns:
        Vector: displacement vector of a displacement gate
    """
    x = math.atleast_1d(x)
    y = math.atleast_1d(y)
    if x.shape[-1] == 1:
        x = math.tile(x, y.shape)
    if y.shape[-1] == 1:
        y = math.tile(y, x.shape)
    return math.sqrt(2 * hbar, dtype=x.dtype) * math.concat([x, y], axis=0)


def beam_splitter_symplectic(theta: Scalar, phi: Scalar) -> Matrix:
    r"""Symplectic matrix of a Beam-splitter gate.

    The dimension is :math:`4\times 4`.

    Args:
        theta: transmissivity parameter
        phi: phase parameter

    Returns:
        Matrix: symplectic (orthogonal) matrix of a beam-splitter gate
    """
    ct = math.cos(theta)
    st = math.sin(theta)
    cp = math.cos(phi)
    sp = math.sin(phi)
    zero = math.zeros_like(theta)
    return math.astensor(
        [
            [ct, -cp * st, zero, -sp * st],
            [cp * st, ct, -sp * st, zero],
            [zero, sp * st, ct, -cp * st],
            [sp * st, zero, cp * st, ct],
        ]
    )


def mz_symplectic(phi_a: Scalar, phi_b: Scalar, internal: bool = False) -> Matrix:
    r"""Symplectic matrix of a Mach-Zehnder gate.

    It supports two conventions:

        * if ``internal=True``, both phases act inside the interferometer:
            ``phi_a`` on the upper arm, ``phi_b`` on the lower arm;
        * if `internal = False` (default), both phases act on the upper arm:
            ``phi_a`` before the first BS, ``phi_b`` after the first BS.

    Args:
        phi_a (float): first phase
        phi_b (float): second phase
        internal (bool): whether phases are in the internal arms (default is False)

    Returns:
        Matrix: symplectic (orthogonal) matrix of a Mach-Zehnder interferometer
    """
    ca = math.cos(phi_a)
    sa = math.sin(phi_a)
    cb = math.cos(phi_b)
    sb = math.sin(phi_b)
    cp = math.cos(phi_a + phi_b)
    sp = math.sin(phi_a + phi_b)

    if internal:
        return 0.5 * math.astensor(
            [
                [ca - cb, -sa - sb, sb - sa, -ca - cb],
                [-sa - sb, cb - ca, -ca - cb, sa - sb],
                [sa - sb, ca + cb, ca - cb, -sa - sb],
                [ca + cb, sb - sa, -sa - sb, cb - ca],
            ]
        )
    else:
        return 0.5 * math.astensor(
            [
                [cp - ca, -sb, sa - sp, -1 - cb],
                [-sa - sp, 1 - cb, -ca - cp, sb],
                [sp - sa, 1 + cb, cp - ca, -sb],
                [cp + ca, -sb, -sa - sp, 1 - cb],
            ]
        )


def two_mode_squeezing_symplectic(r: Scalar, phi: Scalar) -> Matrix:
    r"""Symplectic matrix of a two-mode squeezing gate.

    The dimension is :math:`4\times 4`.

    Args:
        r (float): squeezing magnitude
        phi (float): rotation parameter

    Returns:
        Matrix: symplectic matrix of a two-mode squeezing gate
    """
    cp = math.cos(phi)
    sp = math.sin(phi)
    ch = math.cosh(r)
    sh = math.sinh(r)
    zero = math.zeros_like(r)
    return math.astensor(
        [
            [ch, cp * sh, zero, sp * sh],
            [cp * sh, ch, sp * sh, zero],
            [zero, sp * sh, ch, -cp * sh],
            [sp * sh, zero, -cp * sh, ch],
        ]
    )


def quadratic_phase(s: Scalar):
    r"""Quadratic phase single mode gate.

    .. math::

        P = \exp(i s q^2 / 2 \hbar)

    Reference: https://strawberryfields.ai/photonics/conventions/gates.html

    Args:
        s (float): interaction strength

    Returns:
        Tensor: the :math:`P(s)` matrix (in ``xxpp`` ordering)
    """

    return math.astensor(
        [
            [1, 0],
            [s, 1],
        ]
    )


def controlled_Z(g: Scalar):
    r"""Controlled PHASE gate of two-gaussian modes.

    .. math::

        C_Z = \exp(ig q_1 \otimes q_2 / \hbar).


    Reference: https://arxiv.org/pdf/2110.03247.pdf, Equation 8.
    https://arxiv.org/pdf/1110.3234.pdf, Equation 161.

    Args:
        g (float): interaction strength

    Returns:
        the C_Z(g) matrix (in xxpp ordering)
    """

    return math.astensor(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, g, 1, 0],
            [g, 0, 0, 1],
        ]
    )


def controlled_X(g: Scalar):
    r"""Controlled NOT gate of two-gaussian modes.

    .. math::

        C_X = \exp(ig q_1 \otimes p_2).

    Reference: https://arxiv.org/pdf/2110.03247.pdf, Equation 9.

    Args:
        g (float): interaction strength

    Returns:
        the C_X(g) matrix (in xxpp ordering)
    """

    return math.astensor(
        [
            [1, 0, 0, 0],
            [g, 1, 0, 0],
            [0, 0, 1, -g],
            [0, 0, 0, 1],
        ]
    )


# ~~~~~~~~~~~~~
# CPTP channels
# ~~~~~~~~~~~~~


def CPTP(
    cov: Matrix,
    means: Vector,
    X: Matrix,
    Y: Matrix,
    d: Vector,
    state_modes: Sequence[int],
    transf_modes: Sequence[int],
) -> Tuple[Matrix, Vector]:
    r"""Returns the cov matrix and means vector of a state after undergoing a CPTP channel.

    Computed as ``cov = X \cdot cov \cdot X^T + Y`` and ``d = X \cdot means + d``.

    If the channel is single-mode, ``modes`` can contain ``M`` modes to apply the channel to,
    otherwise it must contain as many modes as the number of modes in the channel.

    Args:
        cov (Matrix): covariance matrix
        means (Vector): means vector
        X (Matrix): the X matrix of the CPTP channel
        Y (Matrix): noise matrix of the CPTP channel
        d (Vector): displacement vector of the CPTP channel
        state_modes (Sequence[int]): modes the state is defined on
        transf_modes (Sequence[int]): modes on which the channel acts
        hbar (float): value of hbar

    Returns:
        Tuple[Matrix, Vector]: the covariance matrix and the means vector of the state after the CPTP channel
    """
    if not set(transf_modes).issubset(state_modes):
        raise ValueError(
            f"The channel should act on a subset of the state modes ({transf_modes} is not a subset of {state_modes})"
        )
    # if single-mode channel, apply to all modes indicated in `modes`
    if X is not None and X.shape[-1] == 2:
        X = math.single_mode_to_multimode_mat(X, len(transf_modes))
    if Y is not None and Y.shape[-1] == 2:
        Y = math.single_mode_to_multimode_mat(Y, len(transf_modes))
    if d is not None and d.shape[-1] == 2:
        d = math.single_mode_to_multimode_vec(d, len(transf_modes))
    indices = [
        state_modes.index(i) for i in transf_modes
    ]  # TODO: do this when calling the method instead of here?
    cov = math.left_matmul_at_modes(X, cov, indices)
    cov = math.right_matmul_at_modes(cov, math.transpose(X), indices)
    cov = math.add_at_modes(cov, Y, indices)
    means = math.matvec_at_modes(X, means, indices)
    means = math.add_at_modes(means, d, indices)
    return cov, means


def loss_XYd(
    transmissivity: Union[Scalar, Vector], nbar: Union[Scalar, Vector], hbar: float
) -> Tuple[Matrix, Matrix, None]:
    r"""Returns the ``X``, ``Y`` matrices and the ``d`` vector for the noisy loss (attenuator) channel.

    .. math::

        X = math.sqrt(gain)
        Y = (gain - 1) * (2 * nbar + 1) * hbar / 2

    Reference: Alessio Serafini - Quantum Continuous Variables (5.77, p. 108)

    Args:
        transmissivity (float): value of the transmissivity, must be between 0 and 1
        nbar (float): photon number expectation value in the environment (0 for pure loss channel)

    Returns:
        Tuple[Matrix, Matrix, None]: the ``X``, ``Y`` matrices and the ``d`` vector for the noisy
        loss channel
    """
    if math.any(transmissivity < 0) or math.any(transmissivity > 1):
        raise ValueError("transmissivity must be between 0 and 1")
    x = math.sqrt(transmissivity)
    X = math.diag(math.concat([x, x], axis=0))
    y = (1 - transmissivity) * (2 * nbar + 1) * hbar / 2
    Y = math.diag(math.concat([y, y], axis=0))
    return X, Y, None


def amp_XYd(gain: Union[Scalar, Vector], nbar: Union[Scalar, Vector], hbar: float) -> Matrix:
    r"""Returns the ``X``, ``Y`` matrices and the d vector for the noisy amplifier channel.

    The quantum limited amplifier channel is recovered for ``nbar = 0.0``.

    Args:
        gain (float): value of the gain > 1
        nbar (float): photon number expectation value in the environment (0 for quantum
            limited amplifier)

    Returns:
        Tuple[Matrix, Vector]: the ``X``, ``Y`` matrices and the ``d`` vector for the noisy
        amplifier channel.
    """
    if math.any(gain < 1):
        raise ValueError("Gain must be larger than 1")
    x = math.sqrt(gain)
    X = math.diag(math.concat([x, x], axis=0))
    y = (gain - 1) * (2 * nbar + 1) * hbar / 2
    Y = math.diag(math.concat([y, y], axis=0))
    return X, Y, None


def noise_Y(noise: Union[Scalar, Vector], hbar: float) -> Matrix:
    r"""Returns the ``X``, ``Y`` matrices and the d vector for the additive noise channel ``(Y = noise * (\hbar / 2) * I)``

    Args:
        noise (float): number of photons in the thermal state

    Returns:
        Tuple[None, Matrix, None]: the ``X``, ``Y`` matrices and the ``d`` vector of the noise channel.
    """
    return math.diag(math.concat([noise, noise], axis=0)) * hbar / 2


def compose_channels_XYd(
    X1: Matrix, Y1: Matrix, d1: Vector, X2: Matrix, Y2: Matrix, d2: Vector
) -> Tuple[Matrix, Matrix, Vector]:
    r"""Returns the combined ``X``, ``Y``, and ``d`` for two CPTP channels.

    Args:
        X1 (Matrix): the ``X`` matrix of the first CPTP channel
        Y1 (Matrix): the ``Y`` matrix of the first CPTP channel
        d1 (Vector): the displacement vector of the first CPTP channel
        X2 (Matrix): the ``X`` matrix of the second CPTP channel
        Y2 (Matrix): the ``Y`` matrix of the second CPTP channel
        d2 (Vector): the displacement vector of the second CPTP channel

    Returns:
        Tuple[Matrix, Matrix, Vector]: the combined ``X``, ``Y``, and ``d`` matrices
    """
    if X1 is None:
        X = X2
    elif X2 is None:
        X = X1
    else:
        X = math.matmul(X2, X1)
    if Y1 is None:
        Y = Y2
    elif Y2 is None:
        Y = Y1
    else:
        Y = math.matmul(math.matmul(X2, Y1), X2) + Y2
    if d1 is None:
        d = d2
    elif d2 is None:
        d = d1
    else:
        d = math.matmul(X2, d1) + d2
    return X, Y, d


# ~~~~~~~~~~~~~~~
# non-TP channels
# ~~~~~~~~~~~~~~~


def general_dyne(
    cov: Matrix,
    means: Vector,
    proj_cov: Matrix,
    proj_means: Vector,
    modes: Sequence[int],
    hbar: float,
) -> Tuple[Scalar, Matrix, Vector]:
    r"""Returns the results of a general dyne measurement.

    Args:
        cov (Matrix): covariance matrix of the state being measured
        means (Vector): means vector of the state being measured
        proj_cov (Matrix): covariance matrix of the state being projected onto
        proj_means (Vector): means vector of the state being projected onto (i.e. the measurement outcome)
        modes (Sequence[int]): modes being measured (modes are indexed from 0 to num_modes-1)

    Returns:
        Tuple[Scalar, Matrix, Vector]: the outcome probability, the post-measurement cov and means vector
    """
    N = cov.shape[-1] // 2
    nB = proj_cov.shape[-1] // 2  # B is the system being measured
    nA = N - nB  # A is the leftover
    Amodes = [i for i in range(N) if i not in modes]
    A, B, AB = partition_cov(cov, Amodes)
    a, b = partition_means(means, Amodes)
    proj_cov = math.cast(proj_cov, B.dtype)
    proj_means = math.cast(proj_means, b.dtype)
    inv = math.inv(B + proj_cov)
    new_cov = A - math.matmul(math.matmul(AB, inv), math.transpose(AB))
    new_means = a + math.matvec(math.matmul(AB, inv), proj_means - b)
    prob = math.exp(-math.sum(math.matvec(inv, proj_means - b) * (proj_means - b))) / (
        pi ** nB * (hbar ** -nB) * math.sqrt(math.det(B + proj_cov))
    )  # TODO: check this (hbar part especially)
    return prob, new_cov, new_means


# ~~~~~~~~~
# utilities
# ~~~~~~~~~
def number_means(cov: Matrix, means: Vector, hbar: float) -> Vector:
    r"""Returns the photon number means vector given a Wigner covariance matrix and a means vector.

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector
        hbar: the value of the Planck constant

    Returns:
        Vector: the photon number means vector
    """
    N = means.shape[-1] // 2
    return (
        means[:N] ** 2
        + means[N:] ** 2
        + math.diag_part(cov[:N, :N])
        + math.diag_part(cov[N:, N:])
        - hbar
    ) / (2 * hbar)


def number_cov(cov: Matrix, means: Vector, hbar: float) -> Matrix:
    r"""Returns the photon number covariance matrix given a Wigner covariance matrix and a means vector.

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector
        hbar: the value of the Planck constant

    Returns:
        Matrix: the photon number covariance matrix
    """
    N = means.shape[-1] // 2
    mCm = cov * means[:, None] * means[None, :]
    dd = math.diag(math.diag_part(mCm[:N, :N] + mCm[N:, N:] + mCm[:N, N:] + mCm[N:, :N])) / (
        2 * hbar ** 2
    )
    CC = (cov ** 2 + mCm) / (2 * hbar ** 2)
    return (
        CC[:N, :N] + CC[N:, N:] + CC[:N, N:] + CC[N:, :N] + dd - 0.25 * math.eye(N, dtype=CC.dtype)
    )


def is_mixed_cov(cov: Matrix) -> bool:  # TODO: deprecate
    r"""Returns ``True`` if the covariance matrix is mixed, ``False`` otherwise."""
    return not is_pure_cov(math.asnumpy(cov), hbar=settings.HBAR)


def trace(cov: Matrix, means: Vector, Bmodes: Sequence[int]) -> Tuple[Matrix, Vector]:
    r"""Returns the covariances and means after discarding the specified modes.

    Args:
        cov (Matrix): covariance matrix
        means (Vector): means vector
        Bmodes (Sequence[int]): modes to discard

    Returns:
        Tuple[Matrix, Vector]: the covariance matrix and the means vector after discarding the specified modes
    """
    N = len(cov) // 2
    Aindices = math.astensor(
        [i for i in range(N) if i not in Bmodes] + [i + N for i in range(N) if i not in Bmodes]
    )
    A_cov_block = math.gather(math.gather(cov, Aindices, axis=0), Aindices, axis=1)
    A_means_vec = math.gather(means, Aindices)
    return A_cov_block, A_means_vec


def partition_cov(cov: Matrix, Amodes: Sequence[int]) -> Tuple[Matrix, Matrix, Matrix]:
    r"""Partitions the covariance matrix into the ``A`` and ``B`` subsystems and the AB coherence block.

    Args:
        cov (Matrix): the covariance matrix
        Amodes (Sequence[int]): the modes of system ``A``

    Returns:
        Tuple[Matrix, Matrix, Matrix]: the cov of ``A``, the cov of ``B`` and the AB block
    """
    N = cov.shape[-1] // 2
    Bindices = math.cast(
        [i for i in range(N) if i not in Amodes] + [i + N for i in range(N) if i not in Amodes],
        "int32",
    )
    Aindices = math.cast(Amodes + [i + N for i in Amodes], "int32")
    A_block = math.gather(math.gather(cov, Aindices, axis=1), Aindices, axis=0)
    B_block = math.gather(math.gather(cov, Bindices, axis=1), Bindices, axis=0)
    AB_block = math.gather(math.gather(cov, Bindices, axis=1), Aindices, axis=0)
    return A_block, B_block, AB_block


def partition_means(means: Vector, Amodes: Sequence[int]) -> Tuple[Vector, Vector]:
    r"""Partitions the means vector into the ``A`` and ``B`` subsystems.

    Args:
        means (Vector): the means vector
        Amodes (Sequence[int]): the modes of system ``A``

    Returns:
        Tuple[Vector, Vector]: the means of ``A`` and the means of ``B``
    """
    N = len(means) // 2
    Bindices = math.cast(
        [i for i in range(N) if i not in Amodes] + [i + N for i in range(N) if i not in Amodes],
        "int32",
    )
    Aindices = math.cast(Amodes + [i + N for i in Amodes], "int32")
    return math.gather(means, Aindices), math.gather(means, Bindices)


def purity(cov: Matrix, hbar: float) -> Scalar:
    r"""Returns the purity of the state with the given covariance matrix.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        float: the purity
    """
    return 1 / math.sqrt(math.det((2 / hbar) * cov))


def symplectic_eigenvals(cov: Matrix, hbar: float) -> Any:
    r"""Returns the sympletic eigenspectrum of a covariance matrix.

    For a pure state, we expect the sympletic eigenvalues to be 1.

    Args:
        cov (Matrix): the covariance matrix
        hbar (float): the value of the Planck constant

    Returns:
        List[float]: the sympletic eigenvalues
    """
    J = math.J(cov.shape[-1] // 2)  # create a sympletic form
    M = math.matmul(1j * J, cov * (2 / hbar))
    vals = math.eigvals(M)  # compute the eigenspectrum
    return math.abs(vals[::2])  # return the even eigenvalues  # TODO: sort?


def von_neumann_entropy(cov: Matrix, hbar: float) -> float:
    r"""Returns the Von Neumann entropy.

    For a pure state, we expect the Von Neumann entropy to be 0.

    Reference: (https://arxiv.org/pdf/1110.3234.pdf), Equations 46-47.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        float: the Von Neumann entropy
    """
    symp_vals = symplectic_eigenvals(cov, hbar)
    g = lambda x: math.xlogy((x + 1) / 2, (x + 1) / 2) - math.xlogy((x - 1) / 2, (x - 1) / 2 + 1e-9)
    entropy = math.sum(g(symp_vals))
    return entropy


def fidelity(
    mu1: Vector, cov1: Matrix, mu2: Vector, cov2: Matrix, hbar=2.0, rtol=1e-05, atol=1e-08
) -> float:
    r"""Returns the fidelity of two gaussian states.

    Reference: `arXiv:2102.05748 <https://arxiv.org/pdf/2102.05748.pdf>`_, equations 95-99.
    Note that we compute the square of equation 98.

    Args:
        mu1 (Vector): the means vector of state 1
        mu2 (Vector): the means vector of state 2
        cov1 (Matrix): the covariance matrix of state 1
        cov1 (Matrix): the covariance matrix of state 2

    Returns:
        float: the fidelity
    """

    cov1 = math.cast(cov1 / hbar, "complex128")  # convert to units where hbar = 1
    cov2 = math.cast(cov2 / hbar, "complex128")  # convert to units where hbar = 1

    mu1 = math.cast(mu1, "complex128")
    mu2 = math.cast(mu2, "complex128")
    deltar = (mu2 - mu1) / math.sqrt(hbar, dtype=mu1.dtype)  # convert to units where hbar = 1
    J = math.J(cov1.shape[0] // 2)
    I = math.eye(cov1.shape[0])
    J = math.cast(J, "complex128")
    I = math.cast(I, "complex128")

    cov12_inv = math.inv(cov1 + cov2)

    V = math.transpose(J) @ cov12_inv @ ((1 / 4) * J + cov2 @ J @ cov1)

    W = -2 * (V @ (1j * J))
    W_inv = math.inv(W)
    matsqrtm = math.sqrtm(
        I - W_inv @ W_inv
    )  # this also handles the case where the input matrix is close to zero
    f0_top = math.det((matsqrtm + I) @ (W @ (1j * J)))
    f0_bot = math.det(cov1 + cov2)

    f0 = (f0_top / f0_bot) ** (1 / 2)  # square of equation 98

    dot = math.sum(
        math.transpose(deltar) * math.matvec(cov12_inv, deltar)
    )  # computing (mu2-mu1)/sqrt(hbar).T @ cov12_inv @ (mu2-mu1)/sqrt(hbar)

    fidelity = f0 * math.exp((-1 / 2) * dot)  # square of equation 95

    return math.cast(fidelity, "float64")


def physical_partial_transpose(cov: Matrix, modes: Sequence[int]) -> Matrix:
    r"""Returns the covariance matrix that corresponds to applying the partial
    transposition on the density matrix of a given set of modes.

    Reference: `https://arxiv.org/abs/quant-ph/9909044 <https://arxiv.org/abs/quant-ph/9909044>`_, Equation 1, 5.

    Args:
        cov (Matrix): the covariance matrix
        modes (Sequence[int]): the modes of system on which transposition is applied

    Returns:
        Matrix: the covariance matrix corresponding to the partially transposed state
    """
    m, _ = cov.shape
    num_modes = m // 2
    mat = [1.0] * m
    for i in modes:
        mat[i + num_modes] = -1.0
    mat = math.astensor(mat, dtype="float64")
    return cov * mat[:, None] * mat[None, :]


def log_negativity(cov: Matrix, hbar: float) -> float:
    r"""Returns the log_negativity of a Gaussian state.

    Reference: `https://arxiv.org/pdf/quant-ph/0102117.pdf <https://arxiv.org/pdf/quant-ph/0102117.pdf>`_ , Equation 57, 61.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        float: the log-negativity
    """
    vals = symplectic_eigenvals(cov, hbar) / (hbar / 2)
    vals_filtered = math.boolean_mask(
        vals, vals < 1.0
    )  # Get rid of terms that would lead to zero contribution.
    if len(vals_filtered) > 0:
        return -math.sum(
            math.log(vals_filtered) / math.cast(math.log(2.0), dtype=vals_filtered.dtype)
        )
    else:
        return 0


def join_covs(covs: Sequence[Matrix]) -> Tuple[Matrix, Vector]:
    r"""Joins the given covariance matrices into a single covariance matrix.

    Args:
        covs (Sequence[Matrix]): the covariance matrices

    Returns:
        Matrix: the joined covariance matrix
    """
    modes = list(range(len(covs[0]) // 2))
    cov = XPMatrix.from_xxpp(covs[0], modes=(modes, modes), like_1=True)
    for i, c in enumerate(covs[1:]):
        modes = list(range(cov.num_modes, cov.num_modes + c.shape[-1] // 2))
        cov = cov @ XPMatrix.from_xxpp(c, modes=(modes, modes), like_1=True)
    return cov.to_xxpp()


def join_means(means: Sequence[Vector]) -> Vector:
    r"""Joins the given means vectors into a single means vector.

    Args:
        means (Sequence[Vector]): the means vectors

    Returns:
        Vector: the joined means vector
    """
    mean = XPVector.from_xxpp(means[0], modes=list(range(len(means[0]) // 2)))
    for i, m in enumerate(means[1:]):
        mean = mean + XPVector.from_xxpp(
            m, modes=list(range(mean.num_modes, mean.num_modes + len(m) // 2))
        )
    return mean.to_xxpp()


def symplectic_inverse(S: Matrix) -> Matrix:
    r"""Returns the inverse of a symplectic matrix.

    Args:
        S (Matrix): the symplectic matrix

    Returns:
        Matrix: the inverse of the symplectic matrix
    """
    S = math.reshape(S, (S.shape[0] // 2, 2, S.shape[1] // 2, 2))
    S = math.transpose(S, (1, 3, 0, 2))
    return math.block(
        [
            [math.transpose(S[1, 1]), -math.transpose(S[0, 1])],
            [-math.transpose(S[1, 0]), math.transpose(S[0, 0])],
        ]
    )


def XYd_dual(X: Matrix, Y: Matrix, d: Vector):
    r"""Returns the dual channel ``(X, Y, d)``.

    Args:
        X (Matrix): the ``X`` matrix
        Y (Matrix): the ``Y`` noise matrix
        d (Vector): the displacement vector

    Returns:
        Tuple[Matrix, Matrix, Vector]: ``(X_dual, Y_dual, d_dual)``
    """
    X_dual = math.inv(X) if X is not None else None
    Y_dual = Y
    d_dual = d
    if Y is not None:
        Y_dual = (
            math.matmul(X_dual, math.matmul(Y, math.transpose(X_dual))) if X_dual is not None else Y
        )
    if d is not None:
        d_dual = math.matvec(X_dual, d) if X_dual is not None else d
    return X_dual, Y_dual, d_dual
